#ifndef DEFORMATION2_HH
#define DEFORMATION2_HH

#include <math.h>

#include <dune/common/exceptions.hh>
#include <dune/grid/geometrygrid.hh>///coordfunction.hh>
//#include <dune/grid/alugrid/3d/entity.hh>
#include <dune/grid/common/datahandleif.hh>
#include <dune/grid/utility/persistentcontainer.hh>
#include <dune/grid/common/mcmgmapper.hh>
//#include <dune/fem/space/common/functionspace.hh>
#include "myparametrization.hh"

#define kronecker(l,m) ((l)==(m)?1:0)


template <class RangeVector, class G, class M, class V>
void computeParametrizations(G& grid, M& mapper, V& container);

template <class V>
V computePointOverBoundary(V& piudue, V& menouno);

template <class V>
V computeDerivatives(V& piu, V& meno);

template <class HostEntity>
HostEntity findFather(const HostEntity & son);

template <class M, class V>
class PMAPExchange;

template <class M, class V>
class FLAGMAPExchange;


// -----------------------------
// Curvilinear Coordinates Cache
// -----------------------------

template <class HostGrid, class Coordinate>
class CoordCache
{
    typedef CoordCache<HostGrid,Coordinate> This;
    static const unsigned int dimension = HostGrid::dimension;
    typedef typename HostGrid::template Codim<dimension>::Entity Vertex;
    typedef typename HostGrid::template Codim<0>::Entity Element;
    typedef typename Dune::template PersistentContainer<HostGrid,Coordinate> DataCache;
    typedef typename Dune::template PersistentContainer<HostGrid,int> NodeFlag;
    typedef typename DataCache::Iterator DCIter;

public:
    explicit CoordCache(const HostGrid& hostGrid)
      : data_(hostGrid,dimension, Dune::FieldVector< double, IDIM >( 0.0 )),
        computedNodes_(hostGrid,dimension) // data_ is a persistent container of vertices!!! codim=dimension!
    {
        //oldMaxLevel_ = hostGrid.maxLevel();
    };

    const Coordinate &operator()(const Element& element, unsigned int corner) const
    {
        return data_(element,corner);
    };

    const Coordinate &operator()(const Vertex& vertex, unsigned int corner) const
    {
        assert(corner==0);
        return data_[vertex];
    };

    template <class Entity>
    const Coordinate &operator()(const Entity& entity, unsigned int corner) const
    {};

    Coordinate &operator()(const Element& element, unsigned int corner)
    {
        return data_(element,corner);
    };
  
    Coordinate &operator()(const Vertex& vertex, unsigned int corner)
    {
        assert(corner==0);
        return data_[vertex];
    };

    template <class Entity>
    Coordinate &operator()(const Entity& entity, unsigned int corner)
    {};

    //TODODEBUG metodi per settare le coordinate dei vertici
    void setVertex(const Vertex & vertex, unsigned int corner, Coordinate position)
    {
        assert(corner==0);
        data_[vertex] = position;
    }

    template <class Entity>
    void setVertex(const Entity & entity, unsigned int corner, Coordinate position)
    {
        //    assert(corner==0);
    }


    void adapt()
    {
        //    data_.update();
        data_.reserve();
    }

    NodeFlag computedNodes_;
private:
    CoordCache(const This &);
    This &operator=(const This &);

    mutable DataCache data_;
};





// ------------------------
// DeformationCoordFunction
// ------------------------

//! deformation depending on a discrete function 
template <class VertexVector, class G, class V, class H, class NEWS>
class DeformationDiscreteFunction2
    : public Dune::DiscreteCoordFunction< double, 3, DeformationDiscreteFunction2< VertexVector, G, V, H, NEWS > > 
{

public:
    typedef Dune::DiscreteCoordFunction< double, 3, DeformationDiscreteFunction2< VertexVector, G, V, H, NEWS > > BaseType;
    typedef VertexVector RangeVector;
    typedef typename V::Iterator::value_type MyParametrizationType;
    typedef typename std::vector<double> ExtremesType;
    typedef typename Dune::FieldVector<double,G::dimension> LocalVector;
    typedef CoordCache<G,RangeVector> Cache;
    typedef typename G::template Codim<0>::Entity::Geometry::LocalCoordinate LC;

    typedef typename Dune::template PersistentContainer<G,RangeVector> HangingNodeCoord;
    typedef typename Dune::template PersistentContainer<G,int> HangingNodeFlag;


    struct LocalVectorComp{
        bool operator()(const LocalVector& lhs, const LocalVector& rhs) const {
            if(lhs[0] < rhs[0])
                return true;
            else if(lhs[0] > rhs[0])
                return false;
            else
                if(lhs[1] < rhs[1])
                    return true;
                else if(lhs[1] > rhs[1])
                    return false;
                else
                    if(lhs[2] < rhs[2])
                        return true;
                    else if (lhs[2] > rhs[2])
                        return false;
                    else
                        return false;
      
      

            // for(int d = 0; d < G::dimension; ++d)
            // 	if(lhs[d] < rhs[d])
        }
    };


    //----------------//
    //  CONSTRUCTORS  //
    //----------------//
  
    //Empty constructor (to be deleted)
    DeformationDiscreteFunction2() {
        parametrizations_=0;
    };

    //Constructor for deformation on macrogrid
    DeformationDiscreteFunction2(G& grid, V& parametrizations, H& helper, NEWS& news) : hostGrid_(grid), parametrizations_(parametrizations), extremes_(grid,0), cache_(grid), helper_(helper), news_(news),hangingCoord_(grid,G::dimension), hangingFlag_(grid,G::dimension) {

        typedef typename G::ctype ct;
        const int dim = G::dimension;
        typedef typename G::LeafGridView LeafGridView;
        typedef typename G::LevelGridView LevelGridView;

        typedef typename LeafGridView::template Codim<0>::template Partition<Dune::All_Partition>::Iterator LeafElementIterator;
        typedef typename LevelGridView::template Codim<0>::template Partition<Dune::All_Partition>::Iterator LevelElementIterator;

        LeafGridView leafview = grid.leafView();
        //Iterators
        LeafElementIterator ebegin = leafview.template begin<0>();
        LeafElementIterator eend = leafview.template end<0>();

        for(LeafElementIterator it = ebegin; it!=eend; ++it){
            extremes_[*it].push_back(0.0);
            extremes_[*it].push_back(1.0);
            extremes_[*it].push_back(0.0);
            extremes_[*it].push_back(1.0);
            extremes_[*it].push_back(0.0);
            extremes_[*it].push_back(1.0);
        }
    
        buildCache();
        // if(grid.comm().rank() == 0){
        //   std::cout << "Parametrizations have been provided by the user" << std::endl;
        // }
        grid.comm().barrier();

        //set faceToVertices and fToI
        // fToI[0][0] = 1;fToI[0][1] = 2;fToI[1][0] = 0;fToI[1][1] = 2;fToI[2][0] = 0;fToI[0][2] = 1;
        faceToVertices[0][0] = 0;    faceToVertices[0][1] = 2;    faceToVertices[0][2] = 4;    faceToVertices[0][3] = 6;
        faceToVertices[1][0] = 1;    faceToVertices[1][1] = 3;    faceToVertices[1][2] = 5;    faceToVertices[1][3] = 7;
        faceToVertices[2][0] = 0;    faceToVertices[2][1] = 1;    faceToVertices[2][2] = 4;    faceToVertices[2][3] = 5;
        faceToVertices[3][0] = 2;    faceToVertices[3][1] = 3;    faceToVertices[3][2] = 6;    faceToVertices[3][3] = 7;
        faceToVertices[4][0] = 0;    faceToVertices[4][1] = 1;    faceToVertices[4][2] = 2;    faceToVertices[4][3] = 3;
        faceToVertices[5][0] = 4;    faceToVertices[5][1] = 5;    faceToVertices[5][2] = 6;    faceToVertices[5][3] = 7;

        //set coordsToVertex;
        LC coord; 
        coord[0] = 0.0;coord[1] = 0.0;coord[2] = 0.0;
        coordsToVertex[coord] = 0;

        coord[0] = 1.0;coord[1] = 0.0;coord[2] = 0.0;
        coordsToVertex[coord] = 1;

        coord[0] = 0.0;coord[1] = 1.0;coord[2] = 0.0;
        coordsToVertex[coord] = 2;

        coord[0] = 1.0;coord[1] = 1.0;coord[2] = 0.0;
        coordsToVertex[coord] = 3;

        coord[0] = 0.0;coord[1] = 0.0;coord[2] = 1.0;
        coordsToVertex[coord] = 4;

        coord[0] = 1.0;coord[1] = 0.0;coord[2] = 1.0;
        coordsToVertex[coord] = 5;

        coord[0] = 0.0;coord[1] = 1.0;coord[2] = 1.0;
        coordsToVertex[coord] = 6;

        coord[0] = 1.0;coord[1] = 1.0;coord[2] = 1.0;
        coordsToVertex[coord] = 7;



        hangingCoord_.resize();
        hangingFlag_.resize(0);

        typedef typename LeafGridView::template Codim<IDIM>::template Partition<Dune::All_Partition>::Iterator LeafVertexIterator;
        typedef typename LeafVertexIterator::Entity LeafVertex;
    
        LeafVertexIterator vbegin = leafview.template begin<IDIM,Dune::All_Partition>();
        LeafVertexIterator vend = leafview.template end<IDIM,Dune::All_Partition>();
        for(LeafVertexIterator vit = vbegin; vit != vend; ++vit){
            const LeafVertex & vtx = *vit;
            hangingCoord_[vtx] = cache_(vtx,0);
        }
    };

    //Constructor for deformation on restart grid
    DeformationDiscreteFunction2(G& grid, V& parametrizations, H& helper, NEWS& news, const std::string & restartFolder, const std::string & extremesFile, const std::string & cacheFile) : hostGrid_(grid), parametrizations_(parametrizations), extremes_(grid,0), cache_(grid), helper_(helper), news_(news),hangingCoord_(grid,G::dimension), hangingFlag_(grid,G::dimension)
    {
        typedef typename G::LeafGridView LeafGridView;
        typedef typename G::LevelGridView LevelGridView;
        typedef typename LevelGridView::template Codim<0>::template Partition<Dune::InteriorBorder_Partition>::Iterator LevelElementIterator;
        typedef typename LevelElementIterator::Entity LevelElement;
        typedef typename LevelElement::Geometry::GlobalCoordinate GC;
        typedef typename LevelElement::Geometry::LocalCoordinate LC;
        typedef typename LeafGridView::template Codim<IDIM>::template Partition<Dune::All_Partition>::Iterator LeafVertexIterator;
        typedef typename LeafVertexIterator::Entity LeafVertex;
        typedef typename Dune::LevelMultipleCodimMultipleGeomTypeMapper< G,Dune::MCMGElementLayout> LevelElementMapperType;
        typedef typename Dune::LeafMultipleCodimMultipleGeomTypeMapper< G,Dune::MCMGVertexLayout> LeafVertexMapperType;


        //IMPORT EXTREMES
        int gridMaxLevel = grid.maxLevel();
        for(int l = 0; l <= gridMaxLevel; ++l)
            {
                LevelGridView levelGridView = grid.levelView(l);
                LevelElementMapperType levelElementMapper(grid,l);
                int nofExtremes = 2*IDIM;
                std::vector<ExtremesType> extremes(levelElementMapper.size(),ExtremesType(nofExtremes,0.0));
                std::stringstream ss;
                ss << restartFolder << "/" << extremesFile << "_level_" << l << "." << helper.rank();
                std::ifstream inEX(ss.str().c_str());
                int indexi;
                while(inEX.good())
                    {
                        inEX >> indexi;
                        for(int i = 0; i < nofExtremes; ++i)
                            inEX >> extremes[indexi][i];
                    }
                inEX.close();

                LevelElementIterator lend = levelGridView.template end<0,Dune::InteriorBorder_Partition>();
                LevelElementIterator lbegin = levelGridView.template begin<0,Dune::InteriorBorder_Partition>();
                for(LevelElementIterator lit = lbegin; lit != lend; ++lit)
                    {
                        const LevelElement & element = *lit;
                        indexi = levelElementMapper.map(element);
                        extremes_[element] = extremes[indexi];
                    }
            }
        MPI_Barrier(helper.getCommunicator());

        //IMPORT VERTEX CACHE
        std::stringstream ss;
        ss << restartFolder << "/" << cacheFile << "." << helper.rank();
        std::ifstream inV(ss.str().c_str());
        LeafVertexMapperType leafVertexMapper(grid);
        int nofVertices = leafVertexMapper.size();
        int nofGridVertex = grid.size(IDIM);
        std::cout << "rank" << helper.rank() << " gridV: " << nofGridVertex << " mapperV " << nofVertices << std::endl;
        //    std::vector<GC> vertices(leafVertexMapper.size(),GC(0.0));
        std::vector<GC> vertices(grid.size(IDIM),GC(0.0));
        int indexi;
        while(inV.good())
            {
                inV >> indexi;
                for(int i = 0; i < IDIM; ++i)
                    inV >> vertices[indexi][i];
            }
        inV.close();
        LeafGridView leafGridView = grid.leafView();
        LeafVertexIterator vend = leafGridView.template end<IDIM,Dune::All_Partition>();
        LeafVertexIterator vbegin = leafGridView.template begin<IDIM,Dune::All_Partition>();
        for(LeafVertexIterator vit = vbegin; vit != vend; ++vit)
            {
                const LeafVertex & element = *vit;
                int indexi2 = leafVertexMapper.map(element);
                if(indexi2 >= leafVertexMapper.size()){
                    std::cout << "rank " << helper.rank() << " vertex: " << indexi2 << std::endl;
                }
                GC coordinate = vertices[indexi2];
                importVertex(coordinate,cache_(element,0));
            }

        //set faceToVertices and fToI
        // fToI[0][0] = 1;fToI[0][1] = 2;fToI[1][0] = 0;fToI[1][1] = 2;fToI[2][0] = 0;fToI[0][2] = 1;
        faceToVertices[0][0] = 0;    faceToVertices[0][1] = 2;    faceToVertices[0][2] = 4;    faceToVertices[0][3] = 6;
        faceToVertices[1][0] = 1;    faceToVertices[1][1] = 3;    faceToVertices[1][2] = 5;    faceToVertices[1][3] = 7;
        faceToVertices[2][0] = 0;    faceToVertices[2][1] = 1;    faceToVertices[2][2] = 4;    faceToVertices[2][3] = 5;
        faceToVertices[3][0] = 2;    faceToVertices[3][1] = 3;    faceToVertices[3][2] = 6;    faceToVertices[3][3] = 7;
        faceToVertices[4][0] = 0;    faceToVertices[4][1] = 1;    faceToVertices[4][2] = 2;    faceToVertices[4][3] = 3;
        faceToVertices[5][0] = 4;    faceToVertices[5][1] = 5;    faceToVertices[5][2] = 6;    faceToVertices[5][3] = 7;

        //set coordsToVertex;
        LC coord; 
        coord[0] = 0.0;coord[1] = 0.0;coord[2] = 0.0;
        coordsToVertex[coord] = 0;

        coord[0] = 1.0;coord[1] = 0.0;coord[2] = 0.0;
        coordsToVertex[coord] = 1;

        coord[0] = 0.0;coord[1] = 1.0;coord[2] = 0.0;
        coordsToVertex[coord] = 2;

        coord[0] = 1.0;coord[1] = 1.0;coord[2] = 0.0;
        coordsToVertex[coord] = 3;

        coord[0] = 0.0;coord[1] = 0.0;coord[2] = 1.0;
        coordsToVertex[coord] = 4;

        coord[0] = 1.0;coord[1] = 0.0;coord[2] = 1.0;
        coordsToVertex[coord] = 5;

        coord[0] = 0.0;coord[1] = 1.0;coord[2] = 1.0;
        coordsToVertex[coord] = 6;

        coord[0] = 1.0;coord[1] = 1.0;coord[2] = 1.0;
        coordsToVertex[coord] = 7;


        hangingCoord_.resize();
        hangingFlag_.resize(0);

        for(LeafVertexIterator vit = vbegin; vit != vend; ++vit){
            const LeafVertex & vtx = *vit;
            hangingCoord_[vtx] = cache_(vtx,0);
        }

    };

    template <class GC>
    void importVertex(GC & vertexCoordinate, RangeVector& y) const
    {
        y=vertexCoordinate;
    };


    template <class HostEntity>
    void calculate(const HostEntity& entity, unsigned int corner,RangeVector& y) const
    {
        //Parametrization is from reference element domain to physical domain.
        //Then, it takes a point (corner) in the reference element domain,
        //it scales this to the domain of the 0-level element of the branch.
        //It computes the physical coordinate of the corner using the parametrization
        //stored at the 0-level element of the branch.
    
        //Ask for the local coordinates of the corner.
        //WARNING: these coordinates are in the reference element associated to the entity,
        //then, in the range [0,1]x[0,1]x[0,1].
        RangeVector center = entity.geometry().center();
        RangeVector x = entity.geometry().corner(corner);
        RangeVector w = entity.geometry().local(x);
        RangeVector scaledW = w;

        //Scale the coordinates to the domain of the 0-level element in the branch
        scaledW[0] = extremes_[entity][0] + (extremes_[entity][1] - extremes_[entity][0]) * w[0];
        scaledW[1] = extremes_[entity][2] + (extremes_[entity][3] - extremes_[entity][2]) * w[1];
        scaledW[2] = extremes_[entity][4] + (extremes_[entity][5] - extremes_[entity][4]) * w[2];

        typename HostEntity::Entity forefather(entity);

        if( entity.hasFather() )
        {
            int entityLevel = entity.level();
            forefather = entity.father();
            for(int i = 1; i < entityLevel; ++i)
                forefather = forefather.father();
        }
        parametrizations_[ forefather ].evaluate(scaledW,y);

    };

    template <class HostEntity>
    void insertEntity(const HostEntity& hostEntity)
    {
        //    RangeVector zeroPosition(0.0);
        typedef typename HostEntity::Geometry HostGeometry;
        const HostGeometry& hostGeo = hostEntity.geometry();
        const unsigned int numCorners = hostGeo.corners();
        for(unsigned int i = 0; i < numCorners; i++){
            //      if(cache_(hostEntity,i)==zeroPosition)
            calculate(hostEntity,i,cache_(hostEntity,i));
        }
    };

    template <class HostEntity, class Mapper>
    void insertEntity(const HostEntity& hostEntity, const Mapper & mapper)
    {
        //    RangeVector zeroPosition(0.0);
        typedef typename HostEntity::Geometry HostGeometry;
        typedef typename G::template Codim<IDIM>::EntityPointer VertexPtr;
        const HostGeometry& hostGeo = hostEntity.geometry();
        const unsigned int numCorners = hostGeo.corners();
        for(unsigned int i = 0; i < numCorners; i++){
            //      if(cache_(hostEntity,i)==zeroPosition)
            VertexPtr v = hostEntity.template subEntity<IDIM>(i);
            int vindex = mapper.map(*v);
            if(computedNodes_[vindex] != 1){
                calculate(hostEntity,i,cache_(hostEntity,i));
                computedNodes_[vindex] = 1;
            }
        }
    };


    //GHOST NODES NOT CACHED!!! FOR THAT CACHING WE NEED COMMUNICATION ON cache_, BECAUSE PARAMETRIZATION CAN NOT BE ACCESSED BY GHOST ELEMENTS
    //PROBABLY WE DON'T NEED GHOST NODES EVALUATION.....
    void buildCache()
    {
        typedef typename G::template Codim<0>::Entity Element;
        typedef typename G::template Codim<0>::template Partition<Dune::InteriorBorder_Partition>::LevelIterator LevelIterator;
        typedef typename G::Traits::HierarchicIterator HierarchicIterator;

        const int maxLevel = hostGrid_.maxLevel();
        const LevelIterator macroEnd = hostGrid_.template lend<0,Dune::InteriorBorder_Partition>(0);
        for(LevelIterator macroIt = hostGrid_.template lbegin<0,Dune::InteriorBorder_Partition>(0); macroIt != macroEnd; ++macroIt)
            {
                const Element& macroElement = *macroIt;
                insertEntity(macroElement);

                const HierarchicIterator hEnd = macroElement.hend(maxLevel);
                for(HierarchicIterator hIt = macroElement.hbegin(maxLevel); hIt != hEnd; ++hIt)
                    insertEntity(*hIt);
            }
    };
  
    template <class NEW>
    void rebuildCache(NEW& isNew)
    {
        int nofElements = 0;
        int nofNews = 0;
        typedef typename G::template Codim<0>::Entity Element;
        typedef typename G::template Codim<0>::template Partition<Dune::InteriorBorder_Partition>::LevelIterator LevelIterator;
        typedef typename G::Traits::HierarchicIterator HierarchicIterator;
        typedef typename Dune::LeafMultipleCodimMultipleGeomTypeMapper< G,Dune::MCMGVertexLayout > LeafVertexMapperType;

        const int maxLevel = hostGrid_.maxLevel();

        //NEWCACHE
        typename G::LeafGridView leafGridView = hostGrid_.leafView();
        int leafNodeSize = leafGridView.size(G::dimension);
        computedNodes_.clear();
        computedNodes_.resize(leafNodeSize,0);
        LeafVertexMapperType leafVertexMapper(hostGrid_);
        //NEWCACHE

        const LevelIterator macroEnd = hostGrid_.template lend<0,Dune::InteriorBorder_Partition>(0);
        for(LevelIterator macroIt = hostGrid_.template lbegin<0,Dune::InteriorBorder_Partition>(0); macroIt != macroEnd; ++macroIt)
            {
                nofElements++;
                const Element& macroElement = *macroIt;
                if(isNew[macroElement]==1){
                    insertEntity(macroElement,leafVertexMapper);
                    nofNews++;
                }
	
                const HierarchicIterator hEnd = macroElement.hend(maxLevel);
                for(HierarchicIterator hIt = macroElement.hbegin(maxLevel); hIt != hEnd; ++hIt){
                    nofElements++;
                    if(isNew[*hIt]==1){
                        insertEntity(*hIt,leafVertexMapper);
                        nofNews++;
                    }
                }
            }
        std::cout << "rank:" << hostGrid_.comm().rank() << " nofNews/nofElements: " << nofNews << "/" << nofElements << std::endl;
    };

    void adaptbak() {
        extremes_.reserve();
        //adatta gli estremi della parallelizzazione
        typedef typename G::ctype ct;
        const int dim = G::dimension;
        typedef typename G::template Codim<0>::template Partition<Dune::All_Partition>::LevelIterator LevelElementIterator;
        typedef typename G::template Codim<0>::EntityPointer EntityPointer;
    
        int maxlevel = hostGrid_.maxLevel();
        Dune::FieldVector<ct,dim> childCenter,fatherCenter,center;
    
        for(int level = 1; level<= maxlevel; ++level){
            //      LevelGridView levelview = hostGrid_.levelView(level);
            LevelElementIterator lbegin = hostGrid_.template lbegin<0,Dune::All_Partition>(level);
            LevelElementIterator lend = hostGrid_.template lend<0,Dune::All_Partition>(level);
            for(LevelElementIterator it = lbegin; it != lend; ++it){
                if((*it).isNew()){
                    EntityPointer father = it->father();
                    fatherCenter = (*father).geometry().center();
                    childCenter = (*it).geometryInFather().center();
                    center = (*it).geometry().center();
                    for(int i=0; i<=2; ++i){
                        if(childCenter[i]==0.25){
                            extremes_[*it].push_back(extremes_[*father][2*i]);
                            extremes_[*it].push_back(extremes_[*father][2*i] + (extremes_[*father][2*i+1] - extremes_[*father][2*i])/2 );
                        }
                        else{
                            extremes_[*it].push_back(extremes_[*father][2*i] + (extremes_[*father][2*i+1] - extremes_[*father][2*i])/2 );
                            extremes_[*it].push_back( extremes_[*father][2*i+1] );
	      
                        }
                    }
                }
            }
        }
    
    }

    void adapt() {
        extremes_.reserve();
        news_.reserve();
        hangingCoord_.resize();
        hangingFlag_.resize(0);
        typename G::CollectiveCommunication commObject = hostGrid_.comm();
        //adatta gli estremi della parallelizzazione
        typedef typename G::ctype ct;
        const int dim = G::dimension;
        typedef typename G::template Codim<0>::template Partition<Dune::InteriorBorder_Partition>::LevelIterator LevelElementIterator;
        typedef typename G::template Codim<0>::EntityPointer EntityPointer;
        typedef typename G::template Codim<0>::Entity Entity;
    
        int maxlevel = hostGrid_.maxLevel();
        Dune::FieldVector<ct,dim> childCenter,fatherCenter,center;
    
        for(int level = 0; level<= maxlevel; ++level){
            //      LevelGridView levelview = hostGrid_.levelView(level);
            LevelElementIterator lbegin = hostGrid_.template lbegin<0,Dune::InteriorBorder_Partition>(level);
            LevelElementIterator lend = hostGrid_.template lend<0,Dune::InteriorBorder_Partition>(level);
            for(LevelElementIterator it = lbegin; it != lend; ++it){
                Entity& element = *it;
                //	if( element.isNew() &&  level > 0){
                if( news_[element] &&  level > 0){
                    childCenter = element.geometryInFather().center();
                    EntityPointer fatherPtr = it->father();
                    Entity& father = *fatherPtr;
                    fatherCenter = father.geometry().center();
                    //center = (*it).geometry().center();
                    extremes_[element].resize( 2 * element.dimension, 0.0 );

                    for(int i=0; i<=2; ++i){
                        if(childCenter[i]==0.25){
                            extremes_[element][2*i] = extremes_[father][2*i];
                            extremes_[element][2*i+1] = extremes_[father][2*i] + (extremes_[father][2*i+1] - extremes_[father][2*i])/2;
                        }
                        else{
                            extremes_[element][2*i] = extremes_[father][2*i] + (extremes_[father][2*i+1] - extremes_[father][2*i])/2;
                            extremes_[element][2*i+1] = extremes_[father][2*i+1];
                        }// end of if(childCenter[i] == 0.25)
                    }// end of for() over childCenter components
                }// end if(element.isNew())
            }// end of loop over level elements
        }// end of loop over levels
        commObject.barrier();
        cache_.adapt();
    }// end of adapt() method



    //-------------------//
    //  EVALUATE METHOD  //
    //-------------------//

    template <class HostEntity> 
    void evaluate ( const HostEntity& entity, 
                    unsigned int corner, 
                    RangeVector &y ) const
    {  

        //Parametrization is from reference element domain to physical domain.
        //Then, it takes a point (corner) in the reference element domain,
        //it scales this to the domain of the 0-level element of the branch.
        //It computes the physical coordinate of the corner using the parametrization
        //stored at the 0-level element of the branch.

        //Ask for the local coordinates of the corner.
        //WARNING: these coordinates are in the reference element associated to the entity,
        //then, in the range [0,1]x[0,1]x[0,1].
        // RangeVector center = entity.geometry().center();
        // RangeVector x = entity.geometry().corner(corner);
        // RangeVector w = entity.geometry().local(x);
        // RangeVector scaledW = w;

        // //Scale the coordinates to the domain of the 0-level element in the branch
        // scaledW[0] = extremes_[entity][0] + (extremes_[entity][1] - extremes_[entity][0]) * w[0];
        // scaledW[1] = extremes_[entity][2] + (extremes_[entity][3] - extremes_[entity][2]) * w[1];
        // scaledW[2] = extremes_[entity][4] + (extremes_[entity][5] - extremes_[entity][4]) * w[2];

        // typename HostEntity::EntityPointer forefather(entity);
        // typename HostEntity::EntityPointer father(entity);

        // if(entity.hasFather()){
        //   int entityLevel = entity.level();
        //   forefather = entity.father();
        //   for(int i = 1; i < entityLevel; ++i){
        // 	forefather = (*forefather).father();
        //   }
        // }
        // parametrizations_[(*forefather)].evaluate(scaledW,y);

        //y=cache_(entity,corner);
        y=hangingCoord_(entity,corner);

        //IDENTITY DEFORMATION
        //y = entity.geometry().corner(corner);

        return;
    };

    template<class Entity>
    RangeVector evaluate(Entity & e, LocalVector & x){
        RangeVector y;
    
        parametrizations_[e].evaluate(x,y);
    
        return y;
    
    }

    void updateParametrizationVectors()
    {
        typedef typename V::iterator gIt;
        typedef typename V::value_type::iterator jIt;
        typedef typename V::value_type::value_type::iterator kIt;
        typedef typename V::value_type::value_type::value_type::iterator iIt;
        typename G::LevelGridView coarseGridView = hostGrid_.levelView(0);
        typename G::LevelGridView::template Codim<0>::template Partition<Dune::InteriorBorder_Partition>::Iterator itend = coarseGridView.template end<0,Dune::InteriorBorder_Partition>();
        typename G::LevelGridView::template Codim<0>::template Partition<Dune::InteriorBorder_Partition>::Iterator itbegin = coarseGridView.template begin<0,Dune::InteriorBorder_Partition>();
        for(typename G::LevelGridView::template Codim<0>::template Partition<Dune::InteriorBorder_Partition>::Iterator it = itbegin; it != itend; ++it)
            {

            }
    };

    void correctJumpHangingNodes(){
    
        typedef typename G::LeafGridView LeafGridView;
        typedef typename LeafGridView::template Codim<0>::template Partition<Dune::InteriorBorder_Partition>::Iterator LeafEntityIterator;
        typedef typename LeafEntityIterator::Entity LeafEntity;
        typedef typename LeafEntity::Geometry EntityGeometry;
        typedef typename EntityGeometry::LocalCoordinate LC;
        typedef typename EntityGeometry::GlobalCoordinate GC;
        typedef typename LeafGridView::IntersectionIterator LeafIntersectionIterator;
        typedef typename LeafIntersectionIterator::Intersection LeafIntersection;
        //    typedef typename LeafIntersection LocalGeometry LocalGeometry;
        typedef typename G::template Codim<0>::EntityPointer EntityPointer;
        typedef typename G::template Codim<1>::EntityPointer FacePointer;
        typedef typename FacePointer::Entity::Geometry FaceGeometry;
        typedef typename G::template Codim<IDIM>::EntityPointer VertexPointer;
        //    typedef typename VertexPointer::Entity::Geometry;

        typedef typename LeafGridView::template Codim<IDIM>::template Partition<Dune::All_Partition>::Iterator LeafVertexIterator;
        typedef typename LeafVertexIterator::Entity LeafVertex;
        typedef typename Dune::LeafMultipleCodimMultipleGeomTypeMapper< G,Dune::MCMGElementLayout> LeafElementMapperType;



        int maxLevel = hostGrid_.maxLevel();
        int globalMaxLevel;
        MPI_Allreduce(&maxLevel,&globalMaxLevel,1,MPI_INT,MPI_MAX,helper_.getCommunicator());

        std::map<LocalVector,int,LocalVectorComp> edgeDeplacementFlag;

        LeafGridView leafGridView = hostGrid_.leafView();
        Dune::MCMGElementLayout<IDIM> layout;
        LeafElementMapperType leafElementMapper(hostGrid_,layout);

        LeafVertexIterator vbegin = leafGridView.template begin<IDIM,Dune::All_Partition>();
        LeafVertexIterator vend = leafGridView.template end<IDIM,Dune::All_Partition>();
        for(LeafVertexIterator vit = vbegin; vit != vend; ++vit){
            const LeafVertex & vtx = *vit;
            hangingCoord_[vtx] = cache_(vtx,0);
            hangingFlag_[vtx] = 0;
        }

        //DEBUG - s
        // LeafElementMapperType leafElementMapper(hostGrid_);
        //DEBUG - f
        LeafEntityIterator itend = leafGridView.template end<0,Dune::InteriorBorder_Partition>();
        LeafEntityIterator itbegin = leafGridView.template begin<0,Dune::InteriorBorder_Partition>();
        for(LeafEntityIterator it = itbegin; it != itend; ++it){
            int level = it->level();
            if(level < globalMaxLevel){
                //DEBUG - s
                // int indexi = leafElementMapper.map(*it);
                //DEBUG - f
                const LeafEntity & e = *it;
                int indexi = leafElementMapper.map(e);
                EntityGeometry eGeo = e.geometry();
                LeafIntersectionIterator iend = leafGridView.iend(e);
                resetEdgeDeplacementFlag(edgeDeplacementFlag);
                //search z direction
                int zDirection = searchDirection(e,2);
                int yDirection = searchDirection(e,1);
                int xDirection = searchDirection(e,0);
                //looppa su intersezioni per individuare hanging da non spostare
                int oldFace = -100;
                std::vector<GC> delta(6);
                for(int f = 0; f < 6; ++f){
                    delta[f][0] = 0.0;
                    delta[f][1] = 0.0;
                    delta[f][2] = 0.0;
                }
                std::vector<bool> moved(6,false);
                for(LeafIntersectionIterator iit = leafGridView.ibegin(e); iit != iend; ++iit){
                    const LeafIntersection & in = *iit;
                    if(in.neighbor()){
                        EntityPointer neighbor = in.outside();
                        int nlevel = neighbor->level();
                        if(level < nlevel){
                            //DEBUG - s
                            // int indexj = leafElementMapper.map(*neighbor);
                            //DEBUG - f
                            int indexInInside = in.indexInInside();
                            if(indexInInside != oldFace){
                                setEdgeDeplacement(edgeDeplacementFlag, indexInInside);
                                oldFace = indexInInside;
                            }
                        }//end l < nl
                    }//end if neighbor
                }//end first loop over intersections
                for(LeafIntersectionIterator iit = leafGridView.ibegin(e); iit != iend; ++iit){
                    const LeafIntersection & in = *iit;
                    if(in.neighbor()){
                        EntityPointer neighbor = in.outside();
                        int nlevel = neighbor->level();
                        int indexj = leafElementMapper.map(*neighbor);
                        if(level < nlevel){
                            int indexInOutside = in.indexInOutside();
                            int indexInInside = in.indexInInside();
                            for(int v = 0; v < 4; ++v){
                                VertexPointer vtxp = neighbor->template subEntity<IDIM>(faceToVertices[indexInOutside][v]);
                                GC vtxGCoord = vtxp->geometry().center();
                                LC vtxLCoord = eGeo.local(vtxGCoord);
                                roundLocalCoord(vtxLCoord);
                                if(hangingFlag_[*vtxp] == 0){
                                    if(edgeDeplacementFlag[vtxLCoord] == 1){
                                        //sposta//
                                        news_[e] = 1;
                                        news_[*neighbor] = 1;
                                        LC p0,p1;
                                        GC gp0,gp1,gpstar,gp1p0;
                                        p0 = p1 = vtxLCoord;
                                        for(int d = 0; d < IDIM; ++d){
                                            if(vtxLCoord[d] < 0.75 && vtxLCoord[d] > 0.25){
                                                p0[d] = 0.0;
                                                p1[d] = 1.0;
                                            }
                                        }
                                        // gp0 = eGeo.global(p0);
                                        // gp1 = eGeo.global(p1);
                                        int p0LIndex = coordsToVertex[p0];
                                        int p1LIndex = coordsToVertex[p1];
                                        // VertexPointer vp0 it->template subEntity<IDIM>(p0LIndex);
                                        // VertexPointer vp1 it->template subEntity<IDIM>(p1LIndex);
                                        gp0=cache_(*it,p0LIndex);
                                        gp1=cache_(*it,p1LIndex);
                                        // gp0 = eGeo.global(p0);
                                        // gp1 = eGeo.global(p1);

                                        // //center projection
                                        // gp1p0 = gp1-gp0;
                                        // //int gp1p0mod = gp1p0.two_norm();
                                        // gp1p0 /= 2;
                                        // gpstar = gp0;
                                        // gpstar.axpy(1.0,gp1p0);
		    

                                        //NEW PROJECTION
                                        gp1p0 = gp1-gp0;
                                        double gp1p0Norm = gp1p0.two_norm();
                                        gp1p0 /= gp1p0Norm;
                                        GC old = cache_(*vtxp,0);
                                        old -= gp0;
                                        double dot = old.dot(gp1p0);
                                        gp1p0 *= dot;
                                        gpstar = gp0+gp1p0;

#if FALSE2D==1
                                        if(vtxLCoord[zDirection] == 1.0 || vtxLCoord[zDirection] == 0.0){
                                            old = cache_(*vtxp,0);
                                            GC temp = (gpstar - old);
                                            temp *= 0.25;
                                            delta[indexInInside] += temp;
                                        }
#endif
                                        hangingCoord_[*vtxp] = gpstar;
                                        //hangingCoord_[*vtxp] = cache_(*vtxp,0);
                                        //hangingFlag_[*vtxp] = 1;
                                    }
                                    else if(edgeDeplacementFlag[vtxLCoord] == 2)
                                        hangingFlag_[*vtxp] = 2;
                                }
                                else if(hangingFlag_[*vtxp] == 1){
                                    news_[e] = 1;
                                    news_[*neighbor] = 1;
                                }
                            }
                        }//end l < nl
                    }//end if neighbor
                }//end intersection

#if FALSE2D == 1
                for(LeafIntersectionIterator iit = leafGridView.ibegin(e); iit != iend; ++iit){
                    const LeafIntersection & in = *iit;
                    if(in.neighbor()){
                        EntityPointer neighbor = in.outside();
                        int nlevel = neighbor->level();
                        if(level < nlevel){
                            int indexInOutside = in.indexInOutside();
                            int indexInInside = in.indexInInside();
                            for(int v = 0; v < 4; ++v){
                                VertexPointer vtxp = neighbor->template subEntity<IDIM>(faceToVertices[indexInOutside][v]);
                                GC vtxGCoord = vtxp->geometry().center();
                                LC vtxLCoord = eGeo.local(vtxGCoord);
                                roundLocalCoord(vtxLCoord);
                                if( ( vtxLCoord[zDirection] == 0.5 && (vtxLCoord[xDirection] == 0.5 || vtxLCoord[yDirection] == 0.5) )&& !moved[indexInInside]){
                                    GC old = cache_(*vtxp,0);
                                    GC gpstar = old + delta[indexInInside];
                                    hangingCoord_[*vtxp] = gpstar;
                                    moved[indexInInside] == true;
                                }
                            }
                        }
                    }
                }
#endif
            }//end l < gml
        }//end element

    };

    void roundLocalCoord(LocalVector& lCoord){
        int size = lCoord.size();
        for(int d = 0; d < size; ++d)
            if(lCoord[d] < 0.25)
                lCoord[d] = 0.0;
            else if(lCoord[d] > 0.75)
                lCoord[d] = 1.0;
            else
                lCoord[d] = 0.5;
    };


    // template<class LC>
    // bool checkJumpHangingNode(int indexInInside, LC localCorner){
    //   int faceType = indexInInside / 2;
    //   if( (localCorner[fToI[faceType][0]] != 0 && localCorner[fToI[faceType][0]] != 1) && (localCorner[fToI[faceType][1]] == 0 || localCorner[fToI[faceType][1]] == 1) ){
    //     return true;
    //   } 
    //   if( (localCorner[fToI[faceType][1]] != 0 && localCorner[fToI[faceType][1]] != 1) && (localCorner[fToI[faceType][0]] == 0 || localCorner[fToI[faceType][0]] == 1) ){
    //     return true;
    //   } 
    
    //   return false;
    // };

    void resetEdgeDeplacementFlag(std::map<LocalVector,int,LocalVectorComp>& edgeDeplacementFlag){
        edgeDeplacementFlag.clear();
        LocalVector edge;
        edge[0] = 0.0; edge[1] = 0.5; edge[2] = 0.0;
        edgeDeplacementFlag[edge] = 0;
        edge[0] = 1.0; edge[1] = 0.5; edge[2] = 0.0;
        edgeDeplacementFlag[edge] = 0;
        edge[0] = 0.5; edge[1] = 0.0; edge[2] = 0.0;
        edgeDeplacementFlag[edge] = 0;
        edge[0] = 0.5; edge[1] = 1.0; edge[2] = 0.0;
        edgeDeplacementFlag[edge] = 0;
        edge[0] = 0.0; edge[1] = 0.0; edge[2] = 0.5;
        edgeDeplacementFlag[edge] = 0;
        edge[0] = 1.0; edge[1] = 0.0; edge[2] = 0.5;
        edgeDeplacementFlag[edge] = 0;
        edge[0] = 0.0; edge[1] = 1.0; edge[2] = 0.5;
        edgeDeplacementFlag[edge] = 0;
        edge[0] = 1.0; edge[1] = 1.0; edge[2] = 0.5;
        edgeDeplacementFlag[edge] = 0;
        edge[0] = 0.0; edge[1] = 0.5; edge[2] = 1.0;
        edgeDeplacementFlag[edge] = 0;
        edge[0] = 1.0; edge[1] = 0.5; edge[2] = 1.0;
        edgeDeplacementFlag[edge] = 0;
        edge[0] = 0.5; edge[1] = 0.0; edge[2] = 1.0;
        edgeDeplacementFlag[edge] = 0;
        edge[0] = 0.5; edge[1] = 1.0; edge[2] = 1.0;
        edgeDeplacementFlag[edge] = 0;

        edge[0] = 1.0; edge[1] = 1.0; edge[2] = 0.5;
        edgeDeplacementFlag[edge] = 0;
    }

    void setEdgeDeplacement(std::map<LocalVector,int,LocalVectorComp>& edgeDeplacementFlag, int indexInInside){
        LocalVector edge;
        switch (indexInInside){
        case (0):
            edge[0] = 0.0; edge[1] = 0.5; edge[2] = 0.0;
            ++edgeDeplacementFlag[edge];
            edge[0] = 0.0; edge[1] = 0.0; edge[2] = 0.5;
            ++edgeDeplacementFlag[edge];
            edge[0] = 0.0; edge[1] = 1.0; edge[2] = 0.5;
            ++edgeDeplacementFlag[edge];
            edge[0] = 0.0; edge[1] = 0.5; edge[2] = 1.0;
            ++edgeDeplacementFlag[edge];
            break;
        case (1):
            edge[0] = 1.0; edge[1] = 0.5; edge[2] = 0.0;
            ++edgeDeplacementFlag[edge];
            edge[0] = 1.0; edge[1] = 0.0; edge[2] = 0.5;
            ++edgeDeplacementFlag[edge];
            edge[0] = 1.0; edge[1] = 1.0; edge[2] = 0.5;
            ++edgeDeplacementFlag[edge];
            edge[0] = 1.0; edge[1] = 0.5; edge[2] = 1.0;
            ++edgeDeplacementFlag[edge];
            break;
        case (2):
            edge[0] = 0.5; edge[1] = 0.0; edge[2] = 0.0;
            ++edgeDeplacementFlag[edge];
            edge[0] = 0.0; edge[1] = 0.0; edge[2] = 0.5;
            ++edgeDeplacementFlag[edge];
            edge[0] = 1.0; edge[1] = 0.0; edge[2] = 0.5;
            ++edgeDeplacementFlag[edge];
            edge[0] = 0.5; edge[1] = 0.0; edge[2] = 1.0;
            ++edgeDeplacementFlag[edge];
            break;
        case (3):
            edge[0] = 0.5; edge[1] = 1.0; edge[2] = 0.0;
            ++edgeDeplacementFlag[edge];
            edge[0] = 0.0; edge[1] = 1.0; edge[2] = 0.5;
            ++edgeDeplacementFlag[edge];
            edge[0] = 1.0; edge[1] = 1.0; edge[2] = 0.5;
            ++edgeDeplacementFlag[edge];
            edge[0] = 0.5; edge[1] = 1.0; edge[2] = 1.0;
            ++edgeDeplacementFlag[edge];
            break;
        case (4):
            edge[0] = 0.0; edge[1] = 0.5; edge[2] = 0.0;
            ++edgeDeplacementFlag[edge];
            edge[0] = 1.0; edge[1] = 0.5; edge[2] = 0.0;
            ++edgeDeplacementFlag[edge];
            edge[0] = 0.5; edge[1] = 0.0; edge[2] = 0.0;
            ++edgeDeplacementFlag[edge];
            edge[0] = 0.5; edge[1] = 1.0; edge[2] = 0.0;
            ++edgeDeplacementFlag[edge];
            break;
        case (5):
            edge[0] = 0.0; edge[1] = 0.5; edge[2] = 1.0;
            ++edgeDeplacementFlag[edge];
            edge[0] = 1.0; edge[1] = 0.5; edge[2] = 1.0;
            ++edgeDeplacementFlag[edge];
            edge[0] = 0.5; edge[1] = 0.0; edge[2] = 1.0;
            ++edgeDeplacementFlag[edge];
            edge[0] = 0.5; edge[1] = 1.0; edge[2] = 1.0;
            ++edgeDeplacementFlag[edge];
            break;
        default:
            std::cout << "wrong intersection index!!" << std::endl;
        }
    };

    template<class E>
    int searchDirection(E & e, int dir){
   
        int direction = -1;
        typename E::Geometry::GlobalCoordinate zero = e.template subEntity<IDIM>(0)->geometry().center();
        std::vector<typename E::Geometry::GlobalCoordinate> vertices(3);
        vertices[0] = e.template subEntity<IDIM>(1)->geometry().center();
        vertices[1] = e.template subEntity<IDIM>(2)->geometry().center();
        vertices[2] = e.template subEntity<IDIM>(4)->geometry().center();

        typename E::Geometry::GlobalCoordinate diff;
        double diffNorm;
        for(int i = 0; i < IDIM; ++i){
            diff = vertices[i] - zero;
            diffNorm = diff.two_norm();
            diff /= diffNorm;
            if(fabs(diff[dir]) > 0.9)
                direction = i;
        }

    
        return direction;
    };
  
    //---------//
    // MEMBERS //
    //---------//
    //int fToI[3][2];
    int faceToVertices[6][4];
    std::map<LC,int,LocalVectorComp>coordsToVertex;
  
    V parametrizations_;
    const G& hostGrid_;
    Dune::PersistentContainer<G,ExtremesType > extremes_;
    Cache cache_;
    H& helper_;
    NEWS& news_;
    std::vector<int> computedNodes_;
    HangingNodeCoord hangingCoord_;
    std::vector<HangingNodeCoord> hangingCoords_;
    HangingNodeFlag hangingFlag_;

    //int oldMaxLevel_;
};//end of deformation class

//------------------------------------------------//
//            PARAMETRIZATION TOOLS               //
//------------------------------------------------//

// IMPORT PARAMETRIZATIONS
template <class Para, class G, class H>
void importParametrizations(G& grid, Para& parametrizations, H& helper, const std::string & restartFolder, const std::string & parametrizationsFile)
{
    typedef typename G::ctype ct;
    const int dim = G::dimension;
    typedef Dune::FieldVector< ct,dim > RangeType;
    typedef typename G::LevelGridView LevelGridView;
    typedef typename LevelGridView::template Codim<0>::template Partition<Dune::InteriorBorder_Partition>::Iterator LevelElementIterator;
    typedef typename LevelElementIterator::Entity LevelElement;
    typedef typename Dune::LevelMultipleCodimMultipleGeomTypeMapper<G,Dune::MCMGElementLayout> LevelElementMapperType;
  
    LevelElementMapperType coarseElementMapper(grid,0);
    std::vector<std::map<int,std::map<int, std::map<int, RangeType> > > > GAMMAS(coarseElementMapper.size());
  
    //OPEN PARAMETRIZATIONS STREAM
    std::stringstream ss;
    ss << restartFolder << "/" << parametrizationsFile << "." << helper.rank();
    std::ifstream inP(ss.str().c_str());
    //IMPORT GAMMA COEFFICIENTS IN A INDEX BASED STRUCTURE
    int indexi;
    double buffer;
    std::string label;
    inP >> label;
    while(inP.good()){
        inP >> indexi;
        for(int g = 0; g <= 3; ++g)
            for(int j = 0; j <= 3; ++j)
                for(int k = 0; k <= 3; ++k)
                    for(int i = 0; i <= 2; ++i){
                        inP >> buffer;
                        GAMMAS[indexi][g][j][k][i]=buffer;
                    }
    }
    inP.close();
    //FILL THE PARAMETRIZATIONS PERSISTENT CONTAINER
    LevelGridView coarseGridView = grid.levelView(0);
    LevelElementIterator lend = coarseGridView.template end<0,Dune::InteriorBorder_Partition>();
    for(LevelElementIterator lit = coarseGridView.template begin<0,Dune::InteriorBorder_Partition>(); lit != lend; ++lit)
        {
            const LevelElement & element = *lit;
            int indexi = coarseElementMapper.map(element);
            MyParametrization<RangeType,RangeType> parametrization(GAMMAS[indexi]);
            parametrizations[element] = parametrization;
        }
};

// FIND FATHER
template <class HostEntity>
HostEntity findFather(const HostEntity & son) {
    return *(son.father());
};

template <class RangeVector, class G, class M, class V>
void computeParametrizations(G& grid, M& mapper, V& container)
{
    const int codim = 0;
    const Dune::PartitionIteratorType ptype = Dune::InteriorBorder_Partition;
    const int dim = G::dimension;
    typedef typename G::ctype ct;
    typedef typename G::LeafGridView LeafGridView;
    typedef typename LeafGridView::template Codim<codim>::template Partition<ptype>::Iterator LeafIterator;
    typedef typename LeafIterator::Entity LeafEntity;
    typedef typename G::template Codim<codim>::EntityPointer EntityPointer;
    typedef typename LeafEntity::Geometry LeafGeometry;
    typedef typename LeafGridView::IntersectionIterator LeafIntersectionIterator;
  
    RangeVector x,w;
  
    //-----------------------------------------------//
    // VERTICES AND DERIVATIVES STRUCTURE DEFINITION //
    //-----------------------------------------------//
    // Vertices and derivatives container. p has 4 indeces and contains 3D FieldVectors,
    // then the fifth index accesses the components of the point or derivative.
    // Index are u,v,w,d: u,v,w \in {-1,0,1,2} they define vertex in reference space,
    // while d \in {0,1,2,3} distinguish between vertices (0) and derivatives.
    // 0 means vertices, 1 derivatives along u, 2 along v and 3 along w.
    std::map<int,std::map<int, std::map<int,std::map<int, RangeVector> > > > pmap;
    std::map<int,std::map<int, std::map<int,bool> > > flagmap;
  
    // PERFORMANCE MAY BE OPTIMIZED USING VECTOR AND A POINTER, OFFSETTING THE INITIAL POSITION
    // USING NEGATIVE INDEX IS LEGAL WITH POINTERS, ACCESSING OUT-OF-BOUNDS IS NOT!!!
  
  
    //TESTVARIABLES
    ct testcomponent;
    int testIntersectionIndex;
    RangeVector testIntersectionCenter,testElementCenter;
    ct buffer;
  
    //-----------------------------------------------//
    // HERMITE THIRD ORDER INTERPOLANTS COEFFICIENTS //
    //-----------------------------------------------//
    // alpha is the structure collecting the coefficients (4 for every polynomials) of the 4 Hermite interpolants.
    // alpha[p][q][j] with p \in {0,1}, q \in {0,1} and j \in {1,2,3,4}:
    // first and second indices are for defining the interpolant
    // (p distinguishes between interpolants used for vertices (edges and faces, too) and the ones used for derivatives)
    // (q distinguishes between interpolants associated to the two extreme values of the parameters {0,1})
    // (j chooses one among 4 components of the )
    std::map<int,std::map<int,std::map<int,double> > > alpha;
    // first interpolant alpha^0_0
    alpha[0][0][0] = 1.0;
    alpha[0][0][1] = 0.0;
    alpha[0][0][2] = -3.0;
    alpha[0][0][3] = 2.0;
    // second interpolant alpha^0_1
    alpha[0][1][0] = 0.0;
    alpha[0][1][1] = 0.0;
    alpha[0][1][2] = 3.0;
    alpha[0][1][3] = -2.0;
    // third interpolant alpha^1_0
    alpha[1][0][0] = 0.0;
    alpha[1][0][1] = 1.0;
    alpha[1][0][2] = -2.0;
    alpha[1][0][3] = 1.0;
    // fourth interpolant alpha^1_1
    alpha[1][1][0] = 0.0;
    alpha[1][1][1] = 0.0;
    alpha[1][1][2] = -1.0;
    alpha[1][1][3] = 1.0;
  
    //---------------------------------------------//
    // COEFFICIENTS OF THE SECOND INTERPOLATION e2 //
    //---------------------------------------------//
    // LAMBDA contains the coefficients of the Q^3 polynomials obtained form the second 3D TFI
    // first 3D TFI is embedded in.
    std::map<int,std::map<int, std::map<int,RangeVector> > >  LAMBDA; 
    std::map<int,std::map<int, std::map<int,RangeVector> > >  GAMMA;
  
    LeafGridView leafGridView = grid.leafView();
  
    // LOOP OVER ELEMENTS (codim 0 entities)
    LeafIterator endit = leafGridView.template end<codim,ptype>();
    for(LeafIterator it = leafGridView.template begin<codim,ptype>(); it != endit; ++it)
        {
            //INITIALIZE p-STRUCTURE to ZERO to USE THE *= OPERATOR
            for(int u=-1; u<=2; ++u){
                for(int v=-1; v<=2; ++v){
                    for(int w=-1; w<=2; ++w){
                        flagmap[u][v][w] = 0;
                        for(int d=0; d<=3; ++d){
                            for(int k=0; k<=2; ++k){
                                pmap[u][v][w][d][k] = 0;
                            }
                        }
                    }
                }
            }

            for(int g=0; g<=3; ++g){
                for(int j=0; j<=3; ++j){
                    for(int k=0; k<=3; ++k){
                        for(int i=0; i<=2; ++i){
                            LAMBDA[g][j][k][i] = 0.0;
                            GAMMA[g][j][k][i] = 0.0;
                        }
                    }
                }
            }
      
            //ELEMENT VERTICES
            pmap[0][0][0][0]=it->geometry().corner(0);
            pmap[1][0][0][0]=it->geometry().corner(1);
            pmap[0][1][0][0]=it->geometry().corner(2);
            pmap[1][1][0][0]=it->geometry().corner(3);
            pmap[0][0][1][0]=it->geometry().corner(4);
            pmap[1][0][1][0]=it->geometry().corner(5);
            pmap[0][1][1][0]=it->geometry().corner(6);
            pmap[1][1][1][0]=it->geometry().corner(7);
      
            testElementCenter = it->geometry().center();
      
            //NEIGHBORS VERTICES
            LeafIntersectionIterator isend = leafGridView.iend(*it);
            for(LeafIntersectionIterator is = leafGridView.ibegin(*it); is != isend; ++is)
                {
                    testIntersectionIndex = is->indexInInside();
                    testIntersectionCenter = is->geometry().center();
                    if((*is).neighbor())
                        {

                            //------------------------------------//
                            // AUXILIARY VERTICES FROM NEIGHBOURS //
                            //------------------------------------//
                            if(testIntersectionIndex == 0)
                                {
                                    pmap[-1][0][0][0] = (*is).outside()->geometry().corner(0);
                                    pmap[-1][1][0][0] = (*is).outside()->geometry().corner(2);
                                    pmap[-1][0][1][0] = (*is).outside()->geometry().corner(4);
                                    pmap[-1][1][1][0] = (*is).outside()->geometry().corner(6);

                                    EntityPointer neighbor = (*is).outside();
                                    LeafIntersectionIterator nisend = leafGridView.iend(*neighbor);
                                    for(LeafIntersectionIterator nis = leafGridView.ibegin(*neighbor); nis != nisend; ++nis){
                                        int nisIndex = nis->indexInInside();
                                        if((*nis).neighbor()){
                                            if(nisIndex != 0 && nisIndex != 1){
                                                if(nisIndex == 2){
                                                    pmap[0][-1][0][0] = (*nis).outside()->geometry().corner(1);
                                                    flagmap[0][-1][0] = 1;
                                                    pmap[0][-1][1][0] = (*nis).outside()->geometry().corner(5);
                                                    flagmap[0][-1][1] = 1;
                                                }
                                                if(nisIndex == 3){
                                                    pmap[0][2][0][0] = (*nis).outside()->geometry().corner(3);
                                                    flagmap[0][2][0] = 1;
                                                    pmap[0][2][1][0] = (*nis).outside()->geometry().corner(7);
                                                    flagmap[0][2][1] = 1;
                                                }
                                                if(nisIndex == 4){
                                                    pmap[0][0][-1][0] = (*nis).outside()->geometry().corner(1);
                                                    flagmap[0][0][-1] = 1;
                                                    pmap[0][1][-1][0] = (*nis).outside()->geometry().corner(3);
                                                    flagmap[0][1][-1] = 1;
                                                }
                                                if(nisIndex == 5){
                                                    pmap[0][0][2][0] = (*nis).outside()->geometry().corner(5);
                                                    flagmap[0][0][2] = 1;
                                                    pmap[0][1][2][0] = (*nis).outside()->geometry().corner(7);
                                                    flagmap[0][1][2] = 1;
                                                }
                                            }
                                        }
                                    }

                                }
                            if(testIntersectionIndex == 1)
                                {
                                    pmap[2][0][0][0] = (*is).outside()->geometry().corner(1);
                                    pmap[2][1][0][0] = (*is).outside()->geometry().corner(3);
                                    pmap[2][0][1][0] = (*is).outside()->geometry().corner(5);
                                    pmap[2][1][1][0] = (*is).outside()->geometry().corner(7);

                                    EntityPointer neighbor = (*is).outside();
                                    LeafIntersectionIterator nisend = leafGridView.iend(*neighbor);
                                    for(LeafIntersectionIterator nis = leafGridView.ibegin(*neighbor); nis != nisend; ++nis){
                                        int nisIndex = nis->indexInInside();
                                        if((*nis).neighbor()){
                                            if(nisIndex != 0 && nisIndex != 1){
                                                if(nisIndex == 2){
                                                    pmap[1][-1][0][0] = (*nis).outside()->geometry().corner(0);
                                                    flagmap[1][-1][0] = 1;
                                                    pmap[1][-1][1][0] = (*nis).outside()->geometry().corner(4);
                                                    flagmap[1][-1][1] = 1;
                                                }
                                                if(nisIndex == 3){
                                                    pmap[1][2][0][0] = (*nis).outside()->geometry().corner(2);
                                                    flagmap[1][2][0] = 1;
                                                    pmap[1][2][1][0] = (*nis).outside()->geometry().corner(6);
                                                    flagmap[1][2][1] = 1;
                                                }
                                                if(nisIndex == 4){
                                                    pmap[1][0][-1][0] = (*nis).outside()->geometry().corner(0);
                                                    flagmap[1][0][-1] = 1;
                                                    pmap[1][1][-1][0] = (*nis).outside()->geometry().corner(2);
                                                    flagmap[1][1][-1] = 1;
                                                }
                                                if(nisIndex == 5){
                                                    pmap[1][0][2][0] = (*nis).outside()->geometry().corner(4);
                                                    flagmap[1][0][2] = 1;
                                                    pmap[1][1][2][0] = (*nis).outside()->geometry().corner(6);
                                                    flagmap[1][1][2] = 1;
                                                }
                                            }
                                        }
                                    }

                                }
                            if(testIntersectionIndex == 2)
                                {
                                    pmap[0][-1][0][0] = (*is).outside()->geometry().corner(0);
                                    pmap[1][-1][0][0] = (*is).outside()->geometry().corner(1);
                                    pmap[0][-1][1][0] = (*is).outside()->geometry().corner(4);
                                    pmap[1][-1][1][0] = (*is).outside()->geometry().corner(5);

                                    EntityPointer neighbor = (*is).outside();
                                    LeafIntersectionIterator nisend = leafGridView.iend(*neighbor);
                                    for(LeafIntersectionIterator nis = leafGridView.ibegin(*neighbor); nis != nisend; ++nis){
                                        int nisIndex = nis->indexInInside();
                                        if((*nis).neighbor()){
                                            if(nisIndex != 2 && nisIndex != 3){
                                                if(nisIndex == 0){
                                                    pmap[-1][0][0][0] = (*nis).outside()->geometry().corner(2);
                                                    flagmap[-1][0][0] = 1;
                                                    pmap[-1][0][1][0] = (*nis).outside()->geometry().corner(6);
                                                    flagmap[-1][0][1] = 1;
                                                }
                                                if(nisIndex == 1){
                                                    pmap[2][0][0][0] = (*nis).outside()->geometry().corner(3);
                                                    flagmap[2][0][0] = 1;
                                                    pmap[2][0][1][0] = (*nis).outside()->geometry().corner(7);
                                                    flagmap[2][0][1] = 1;
                                                }
                                                if(nisIndex == 4){
                                                    pmap[0][0][-1][0] = (*nis).outside()->geometry().corner(2);
                                                    flagmap[0][0][-1] = 1;
                                                    pmap[1][0][-1][0] = (*nis).outside()->geometry().corner(3);
                                                    flagmap[1][0][-1] = 1;
                                                }
                                                if(nisIndex == 5){
                                                    pmap[0][0][2][0] = (*nis).outside()->geometry().corner(2);
                                                    flagmap[0][0][2] = 1;
                                                    pmap[1][0][2][0] = (*nis).outside()->geometry().corner(7);
                                                    flagmap[1][0][2] = 1;
                                                }
                                            }
                                        }
                                    }

                                }
                            if(testIntersectionIndex == 3)
                                {
                                    pmap[0][2][0][0] = (*is).outside()->geometry().corner(2);
                                    pmap[1][2][0][0] = (*is).outside()->geometry().corner(3);
                                    pmap[0][2][1][0] = (*is).outside()->geometry().corner(6);
                                    pmap[1][2][1][0] = (*is).outside()->geometry().corner(7);

                                    EntityPointer neighbor = (*is).outside();
                                    LeafIntersectionIterator nisend = leafGridView.iend(*neighbor);
                                    for(LeafIntersectionIterator nis = leafGridView.ibegin(*neighbor); nis != nisend; ++nis){
                                        int nisIndex = nis->indexInInside();
                                        if((*nis).neighbor()){
                                            if(nisIndex != 2 && nisIndex != 3){
                                                if(nisIndex == 0){
                                                    pmap[-1][1][0][0] = (*nis).outside()->geometry().corner(0);
                                                    flagmap[-1][1][0] = 1;
                                                    pmap[-1][1][1][0] = (*nis).outside()->geometry().corner(4);
                                                    flagmap[-1][1][1] = 1;
                                                }
                                                if(nisIndex == 1){
                                                    pmap[2][1][0][0] = (*nis).outside()->geometry().corner(1);
                                                    flagmap[2][1][0] = 1;
                                                    pmap[2][1][1][0] = (*nis).outside()->geometry().corner(5);
                                                    flagmap[2][1][1] = 1;
                                                }
                                                if(nisIndex == 4){
                                                    pmap[0][1][-1][0] = (*nis).outside()->geometry().corner(0);
                                                    flagmap[0][1][-1] = 1;
                                                    pmap[1][1][-1][0] = (*nis).outside()->geometry().corner(1);
                                                    flagmap[1][1][-1] = 1;
                                                }
                                                if(nisIndex == 5){
                                                    pmap[0][1][2][0] = (*nis).outside()->geometry().corner(4);
                                                    flagmap[0][1][2] = 1;
                                                    pmap[1][1][2][0] = (*nis).outside()->geometry().corner(5);
                                                    flagmap[1][1][2] = 1;
                                                }
                                            }
                                        }
                                    }

                                }
                            if(testIntersectionIndex == 4)
                                {
                                    pmap[0][0][-1][0] = (*is).outside()->geometry().corner(0);
                                    pmap[1][0][-1][0] = (*is).outside()->geometry().corner(1);
                                    pmap[0][1][-1][0] = (*is).outside()->geometry().corner(2);
                                    pmap[1][1][-1][0] = (*is).outside()->geometry().corner(3);

                                    EntityPointer neighbor = (*is).outside();
                                    LeafIntersectionIterator nisend = leafGridView.iend(*neighbor);
                                    for(LeafIntersectionIterator nis = leafGridView.ibegin(*neighbor); nis != nisend; ++nis){
                                        int nisIndex = nis->indexInInside();
                                        if((*nis).neighbor()){
                                            if(nisIndex != 4 && nisIndex != 5){
                                                if(nisIndex == 0){
                                                    pmap[-1][0][0][0] = (*nis).outside()->geometry().corner(4);
                                                    flagmap[-1][0][0] = 1;
                                                    pmap[-1][1][0][0] = (*nis).outside()->geometry().corner(6);
                                                    flagmap[-1][1][0] = 1;
                                                }
                                                if(nisIndex == 1){
                                                    pmap[2][0][0][0] = (*nis).outside()->geometry().corner(5);
                                                    flagmap[2][0][0] = 1;
                                                    pmap[2][1][0][0] = (*nis).outside()->geometry().corner(7);
                                                    flagmap[2][1][0] = 1;
                                                }
                                                if(nisIndex == 2){
                                                    pmap[0][-1][0][0] = (*nis).outside()->geometry().corner(4);
                                                    flagmap[0][-1][1] = 1;
                                                    pmap[1][-1][0][0] = (*nis).outside()->geometry().corner(5);
                                                    flagmap[1][-1][0] = 1;
                                                }
                                                if(nisIndex == 3){
                                                    pmap[0][2][0][0] = (*nis).outside()->geometry().corner(6);
                                                    flagmap[0][2][0] = 1;
                                                    pmap[1][2][0][0] = (*nis).outside()->geometry().corner(7);
                                                    flagmap[1][2][0] = 1;
                                                }
                                            }
                                        }
                                    }


                                }
                            if(testIntersectionIndex == 5)
                                {
                                    pmap[0][0][2][0] = (*is).outside()->geometry().corner(4);
                                    pmap[1][0][2][0] = (*is).outside()->geometry().corner(5);
                                    pmap[0][1][2][0] = (*is).outside()->geometry().corner(6);
                                    pmap[1][1][2][0] = (*is).outside()->geometry().corner(7);

                                    EntityPointer neighbor = (*is).outside();
                                    LeafIntersectionIterator nisend = leafGridView.iend(*neighbor);
                                    for(LeafIntersectionIterator nis = leafGridView.ibegin(*neighbor); nis != nisend; ++nis){
                                        int nisIndex = nis->indexInInside();
                                        if((*nis).neighbor()){
                                            if(nisIndex != 4 && nisIndex != 5){
                                                if(nisIndex == 0){
                                                    pmap[-1][0][1][0] = (*nis).outside()->geometry().corner(0);
                                                    flagmap[-1][0][1] = 1;
                                                    pmap[-1][1][1][0] = (*nis).outside()->geometry().corner(2);
                                                    flagmap[-1][1][1] = 1;
                                                }
                                                if(nisIndex == 1){
                                                    pmap[2][0][1][0] = (*nis).outside()->geometry().corner(1);
                                                    flagmap[2][0][1] = 1;
                                                    pmap[2][1][1][0] = (*nis).outside()->geometry().corner(3);
                                                    flagmap[2][1][1] = 1;
                                                }
                                                if(nisIndex == 2){
                                                    pmap[0][-1][1][0] = (*nis).outside()->geometry().corner(0);
                                                    flagmap[0][-1][1] = 1;
                                                    pmap[1][-1][1][0] = (*nis).outside()->geometry().corner(1);
                                                    flagmap[1][-1][1] = 1;
                                                }
                                                if(nisIndex == 3){
                                                    pmap[0][2][1][0] = (*nis).outside()->geometry().corner(2);
                                                    flagmap[0][2][1] = 1;
                                                    pmap[1][2][1][0] = (*nis).outside()->geometry().corner(3);
                                                    flagmap[1][2][1] = 1;
                                                }
                                            }
                                        }
                                    }

                                }
                        }
                    else
                        {
                            // NO NEIGHBOR AVAILABLE
                            //------------------------------------------------------------------------------------//
                            // AUXILIARY VERTICES COMPUTED TO PRESERVE THE DERIVATIVE OF THE OF THE LINEAR SEGMENT //
                            //------------------------------------------------------------------------------------//
                            if(testIntersectionIndex == 0)
                                {
                                    if(!flagmap[-1][0][0])
                                        pmap[-1][0][0][0] = computePointOverBoundary(pmap[0][0][0][0],pmap[1][0][0][0]);
                                    if(!flagmap[-1][1][0])
                                        pmap[-1][1][0][0] = computePointOverBoundary(pmap[0][1][0][0],pmap[1][1][0][0]);
                                    if(!flagmap[-1][0][1])
                                        pmap[-1][0][1][0] = computePointOverBoundary(pmap[0][0][1][0],pmap[1][0][1][0]);
                                    if(!flagmap[-1][1][1])
                                        pmap[-1][1][1][0] = computePointOverBoundary(pmap[0][1][1][0],pmap[1][1][1][0]);
                                }
                            if(testIntersectionIndex == 1)
                                {
                                    if(!flagmap[2][0][0])
                                        pmap[2][0][0][0] = computePointOverBoundary(pmap[1][0][0][0],pmap[0][0][0][0]);
                                    if(!flagmap[2][1][0])
                                        pmap[2][1][0][0] = computePointOverBoundary(pmap[1][1][0][0],pmap[0][1][0][0]);
                                    if(!flagmap[2][0][1])
                                        pmap[2][0][1][0] = computePointOverBoundary(pmap[1][0][1][0],pmap[0][0][1][0]);
                                    if(!flagmap[2][1][1])
                                        pmap[2][1][1][0] = computePointOverBoundary(pmap[1][1][1][0],pmap[0][1][1][0]);
                                }
                            if(testIntersectionIndex == 2)
                                {
                                    if(!flagmap[0][-1][0])
                                        pmap[0][-1][0][0] = computePointOverBoundary(pmap[0][0][0][0],pmap[0][1][0][0]);
                                    if(!flagmap[1][-1][0])
                                        pmap[1][-1][0][0] = computePointOverBoundary(pmap[1][0][0][0],pmap[1][1][0][0]);
                                    if(!flagmap[0][-1][1])
                                        pmap[0][-1][1][0] = computePointOverBoundary(pmap[0][0][1][0],pmap[0][1][1][0]);
                                    if(!flagmap[1][-1][1])
                                        pmap[1][-1][1][0] = computePointOverBoundary(pmap[1][0][1][0],pmap[1][1][1][0]);
                                }
                            if(testIntersectionIndex == 3)
                                {
                                    if(!flagmap[0][2][0])
                                        pmap[0][2][0][0] = computePointOverBoundary(pmap[0][1][0][0],pmap[0][0][0][0]);
                                    if(!flagmap[1][2][0])
                                        pmap[1][2][0][0] = computePointOverBoundary(pmap[1][1][0][0],pmap[1][0][0][0]);
                                    if(!flagmap[0][2][1])
                                        pmap[0][2][1][0] = computePointOverBoundary(pmap[0][1][1][0],pmap[0][0][1][0]);
                                    if(!flagmap[1][2][1])
                                        pmap[1][2][1][0] = computePointOverBoundary(pmap[1][1][1][0],pmap[1][0][1][0]);
                                }
                            if(testIntersectionIndex == 4)
                                {
                                    if(!flagmap[0][0][-1])
                                        pmap[0][0][-1][0] = computePointOverBoundary(pmap[0][0][0][0],pmap[0][0][1][0]);
                                    if(!flagmap[1][0][-1])
                                        pmap[1][0][-1][0] = computePointOverBoundary(pmap[1][0][0][0],pmap[1][0][1][0]);
                                    if(!flagmap[0][1][-1])
                                        pmap[0][1][-1][0] = computePointOverBoundary(pmap[0][1][0][0],pmap[0][1][1][0]);
                                    if(!flagmap[1][1][-1])
                                        pmap[1][1][-1][0] = computePointOverBoundary(pmap[1][1][0][0],pmap[1][1][1][0]);
                                }
                            if(testIntersectionIndex == 5)
                                {
                                    if(!flagmap[0][0][2])
                                        pmap[0][0][2][0] = computePointOverBoundary(pmap[0][0][1][0],pmap[0][0][0][0]);
                                    if(!flagmap[1][0][2])
                                        pmap[1][0][2][0] = computePointOverBoundary(pmap[1][0][1][0],pmap[1][0][0][0]);
                                    if(!flagmap[0][1][2])
                                        pmap[0][1][2][0] = computePointOverBoundary(pmap[0][1][1][0],pmap[0][1][0][0]);
                                    if(!flagmap[1][1][2])
                                        pmap[1][1][2][0] = computePointOverBoundary(pmap[1][1][1][0],pmap[1][1][0][0]);
                                }
                        }
                    testIntersectionIndex = 0;
                }
      
            //-------------//
            // DERIVATIVES //
            //-------------//
            // along u - pmap[u][v][w][1] with u,v,w \in {0,1}
            // along v - pmap[u][v][w][2] with u,v,w \in {0,1}
            // along w - pmap[u][v][w][3] with u,v,w \in {0,1}
            for(int u=0; u<=1; ++u){
                for(int v=0; v<=1; ++v){
                    for(int w=0; w<=1; ++w){
                        pmap[u][v][w][1] = computeDerivatives(pmap[u+1][v][w][0],pmap[u-1][v][w][0]);
                        pmap[u][v][w][2] = computeDerivatives(pmap[u][v+1][w][0],pmap[u][v-1][w][0]);
                        pmap[u][v][w][3] = computeDerivatives(pmap[u][v][w+1][0],pmap[u][v][w-1][0]);
                    }
                }
            }
      
            //-------------------------------------------------//
            // COMPUTE LAMBDA COEFFICIENTS - 2nd INTERPOLATION //
            //-------------------------------------------------//
            for(int g = 0; g <= 3; ++g){
                for(int j = 0; j <= 3; ++j){
                    for(int k = 0; k <= 3; ++k){
                        for(int i = 0; i <= 2; ++i){
	      
                            for(int b = 0; b <= 1; ++b){
                                for(int l = 0; l <= 1; ++l){
                                    for(int m = 0; m <= 1; ++m){
		    
                                        for(int p = 0; p <= 1; ++p){
                                            for(int q = 0; q <= 1; ++q){
                                                if(p==0){
                                                    buffer = (pmap[b][q][l][0][i] * alpha[p][q][j] * alpha[0][l][k] * alpha[0][b][g] + pmap[b][m][q][0][i] * alpha[p][q][k] * alpha[0][m][j] * alpha[0][b][g]) * kronecker(l,m);
                                                }
                                                else if(p==1){
                                                    buffer = (pmap[b][q][l][2][i] * alpha[p][q][j] * alpha[0][l][k] * alpha[0][b][g] + pmap[b][m][q][3][i] * alpha[p][q][k] * alpha[0][m][j] * alpha[0][b][g]) * kronecker(l,m);
                                                }
                                                LAMBDA[g][j][k][i] += buffer;
                                            } // closing q-loop
                                        } // closing p-loop
                                        buffer = -pmap[b][l][m][0][i] * alpha[0][l][j] * alpha[0][m][k] * alpha[0][b][g];
                                        LAMBDA[g][j][k][i] += buffer;
		    
                                        for(int p = 0; p <= 1; ++p){
                                            for(int q = 0; q <= 1; ++q){
                                                if(p==0){
                                                    buffer = (pmap[q][b][l][0][i] * alpha[p][q][g] * alpha[0][l][k] * alpha[0][b][j] + pmap[m][b][q][0][i] * alpha[p][q][k] * alpha[0][m][g] * alpha[0][b][j]) * kronecker(l,m);
                                                }
                                                else if(p==1){
                                                    buffer = (pmap[q][b][l][1][i] * alpha[p][q][g] * alpha[0][l][k] * alpha[0][b][j] + pmap[m][b][q][3][i] * alpha[p][q][k] * alpha[0][m][g] * alpha[0][b][j]) * kronecker(l,m);
                                                }
                                                LAMBDA[g][j][k][i] += buffer;
                                            } // closing q-loop
                                        } // closing p-loop
                                        buffer = -pmap[l][b][m][0][i] * alpha[0][l][g] * alpha[0][m][k] * alpha[0][b][j];
                                        LAMBDA[g][j][k][i] += buffer;
		    
                                    } // closing m-loop
                                } // closing l-loop
                            } // closing b-loop
	      
                            for(int s = 0; s <= 3; ++s){
                                for(int b = 0; b <= 1; ++b){
                                    for(int a = 0; a <= 1; ++a){
                                        for(int l = 0; l <= 1; ++l){
                                            for(int m = 0; m <= 1; ++m){
			
                                                for(int p = 0; p <= 1; ++p){
                                                    for(int q = 0; q <= 1; ++q){
                                                        if(p==0){
                                                            buffer = (-pmap[a][q][l][0][i] * alpha[p][q][s] * alpha[0][l][k] * alpha[0][a][g] * alpha[0][b][j] - pmap[a][m][q][0][i] * alpha[p][q][k] * alpha[0][m][s] * alpha[0][a][g] * alpha[0][b][j]) * kronecker(l,m) * Pow(b,s);
                                                        }
                                                        else if(p==1){
                                                            buffer = (-pmap[a][q][l][2][i] * alpha[p][q][s] * alpha[0][l][k] * alpha[0][a][g] * alpha[0][b][j] - pmap[a][m][q][3][i] * alpha[p][q][k] * alpha[0][m][s] * alpha[0][a][g] * alpha[0][b][j]) * kronecker(l,m) * Pow(b,s);
                                                        }
                                                        LAMBDA[g][j][k][i] += buffer;
                                                    } // closing q-loop
                                                } // closing p-loop
                                                buffer = pmap[a][l][m][0][i] * alpha[0][l][s] * alpha[0][m][k] * alpha[0][a][g] * alpha[0][b][j] * Pow(b,s);
                                                LAMBDA[g][j][k][i] += buffer;
			
                                            } // closing m-loop
                                        } // closing l-loop
                                    } // closing a-loop
                                } // closing b-loop
                            } // closing s-loop
	      
                        } // closing i-loop
                    } // closing k-loop
                } // closing j-loop
            }; // closing g-loop
      
            //------------------------------------------------------------------------------------//
            // COMPUTE GAMMA COEFFICIENTS - 3rd AND FINAL  INTERPOLATION //
            //------------------------------------------------------------------------------------//
      
            for(int g = 0; g <= 3; ++g){
                for(int j = 0; j <= 3; ++j){
                    for(int k = 0; k <= 3; ++k){
                        for(int i = 0; i <= 2; ++i){
	      
                            GAMMA[g][j][k][i] = LAMBDA[g][j][k][i];
	      
                            for(int c = 0; c <= 1; ++c){
                                for(int l = 0; l <= 1; ++l){
                                    for(int m = 0; m <= 1; ++m){
		    
                                        for(int p = 0; p <= 1; ++p){
                                            for(int q = 0; q <= 1; ++q){
                                                if(p==0){
                                                    buffer = ( pmap[q][l][c][0][i] * alpha[p][q][g] * alpha[0][l][j] * alpha[0][c][k] + pmap[m][q][c][0][i] * alpha[p][q][j] * alpha[0][m][g] * alpha[0][c][k]) * kronecker(l,m);
                                                    GAMMA[g][j][k][i] += buffer;
                                                }
                                                else if(p==1){
                                                    buffer = ( pmap[q][l][c][1][i] * alpha[p][q][g] * alpha[0][l][j] * alpha[0][c][k] + pmap[m][q][c][2][i] * alpha[p][q][j] * alpha[0][m][g] * alpha[0][c][k]) * kronecker(l,m);
                                                    GAMMA[g][j][k][i] += buffer;
                                                }
                                            }
                                        }
                                        buffer = -pmap[l][m][c][0][i] * alpha[0][l][g] * alpha[0][m][j] * alpha[0][c][k];
                                        GAMMA[g][j][k][i] += buffer;
                                    }
                                }
                            }
	      
                            for(int h = 0; h <= 3; ++h){
                                for(int c = 0; c <= 1; ++c){
                                    buffer = -alpha[0][c][k] * LAMBDA[g][j][h][i] * Pow(c,h);
                                    GAMMA[g][j][k][i] += buffer;
                                }
                            }
	      
                        }
                    }
                }
            }
      
            //-------------------------------------------------------------------------------//
            // BUILDING ELEMENT PARAMETRIZATION AND STORING IT IN PARAMETRIZATIONS ATTRIBUTE //
            //-------------------------------------------------------------------------------//
            MyParametrization<RangeVector,RangeVector> testparametrization(GAMMA);
            container[*it] = testparametrization;// se il container è un persistentcontainer
            //container[mapper.map(*it)] = testparametrization; // se il container è un vector
            RangeVector center = it->geometry().center();
            //      std::cout << "Element center " << center << std::endl;
        }
    return;

};

template <class V>
V computePointOverBoundary(V& piudue, V& menouno)
{
    V p;
    p = piudue;
    p *= 2;
    p -= menouno;
    return p;
};


template <class V>
V computeDerivatives(V& piu, V& meno)
{
    V p;
    p = piu;
    p -= meno;
    p *= 0.5;
    return p;
};





//NEW COMPUTEPARAMETRIZATIONS

template <class VEC>
bool derivativeModuleEqual(VEC a1, VEC a2){

    size_t a1size = a1.size();
    size_t a2size = a2.size();
    assert(a1size==a2size);
  
    for(size_t i = 0; i < a1size; ++i){
        a1[i] = fabs(a1[i]);
        a2[i] = fabs(a2[i]);
        if(fabs(a1[i]-a2[i]) > 1.0e-8)
            return false;
    }

    return true;
};

template <class RangeVector, class GP, class M, class V>
void computeParametrizationsUNSTRUCTURED(GP& gridPtr, M& mapper, V& container)
{
    const int codim = 0;
    const Dune::PartitionIteratorType ptype = Dune::InteriorBorder_Partition;
    typedef typename GP::Grid G;
    const int dim = G::dimension;
    typedef typename G::ctype ct;
    typedef typename G::LeafGridView LeafGridView;
    typedef typename LeafGridView::template Codim<codim>::template Partition<ptype>::Iterator EIterator;
    typedef typename EIterator::Entity Element;
    typedef typename Element::Geometry EGeometry;
    typedef typename EGeometry::GlobalCoordinate GC;
    typedef typename EGeometry::LocalCoordinate LC;
    typedef typename LeafGridView::IntersectionIterator IIterator;
    typedef typename IIterator::Intersection Intersection;
    typedef typename Intersection::LocalGeometry ILocalGeometry;
    typedef typename Intersection::Geometry IGeometry;
    typedef typename IGeometry::GlobalCoordinate IGC;
    typedef typename ILocalGeometry::LocalCoordinate ILC;
  

    RangeVector x,w;
    G& grid = *gridPtr;
  
    //-----------------------------------------------//
    // VERTICES AND DERIVATIVES STRUCTURE DEFINITION //
    //-----------------------------------------------//
    // Vertices and derivatives container. p has 4 indeces and contains 3D FieldVectors,
    // then the fifth index accesses the components of the point or derivative.
    // Index are u,v,w,d: u,v,w \in {-1,0,1,2} they define vertex in reference space,
    // while d \in {0,1,2,3} distinguish between vertices (0) and derivatives.
    // 0 means vertices, 1 derivatives along u, 2 along v and 3 along w.
    typedef typename std::map<int,std::map<int, std::map<int,std::map<int, RangeVector> > > > PmapType;
    std::vector< PmapType> pmap( mapper.size() );
    typedef typename std::map<int,std::map<int, std::map<int,bool> > > FlagmapType;
    std::vector< FlagmapType > flagmap( mapper.size() );
  
    // PERFORMANCE MAY BE OPTIMIZED USING VECTOR AND A POINTER, OFFSETTING THE INITIAL POSITION
    // USING NEGATIVE INDEX IS LEGAL WITH POINTERS, ACCESSING OUT-OF-BOUNDS IS NOT!!!
  
  
    //TESTVARIABLES
    ct testcomponent;
    int testIntersectionIndex;
    RangeVector testIntersectionCenter,testElementCenter;
    ct buffer;
  
    //-----------------------------------------------//
    // HERMITE THIRD ORDER INTERPOLANTS COEFFICIENTS //
    //-----------------------------------------------//
    // alpha is the structure collecting the coefficients (4 for every polynomials) of the 4 Hermite interpolants.
    // alpha[p][q][j] with p \in {0,1}, q \in {0,1} and j \in {1,2,3,4}:
    // first and second indices are for defining the interpolant
    // (p distinguishes between interpolants used for vertices (edges and faces, too) and the ones used for derivatives)
    // (q distinguishes between interpolants associated to the two extreme values of the parameters {0,1})
    // (j chooses one among 4 components of the )
    std::map<int,std::map<int,std::map<int,double> > > alpha;
    // first interpolant alpha^0_0
    alpha[0][0][0] = 1.0;
    alpha[0][0][1] = 0.0;
    alpha[0][0][2] = -3.0;
    alpha[0][0][3] = 2.0;
    // second interpolant alpha^0_1
    alpha[0][1][0] = 0.0;
    alpha[0][1][1] = 0.0;
    alpha[0][1][2] = 3.0;
    alpha[0][1][3] = -2.0;
    // third interpolant alpha^1_0
    alpha[1][0][0] = 0.0;
    alpha[1][0][1] = 1.0;
    alpha[1][0][2] = -2.0;
    alpha[1][0][3] = 1.0;
    // fourth interpolant alpha^1_1
    alpha[1][1][0] = 0.0;
    alpha[1][1][1] = 0.0;
    alpha[1][1][2] = -1.0;
    alpha[1][1][3] = 1.0;
  
    //---------------------------------------------//
    // COEFFICIENTS OF THE SECOND INTERPOLATION e2 //
    //---------------------------------------------//
    // LAMBDA contains the coefficients of the Q^3 polynomials obtained form the second 3D TFI
    // first 3D TFI is embedded in.
    std::map<int,std::map<int, std::map<int,RangeVector> > >  LAMBDA; 
    std::map<int,std::map<int, std::map<int,RangeVector> > >  GAMMA;
    for(int g=0; g<=3; ++g){
        for(int j=0; j<=3; ++j){
            for(int k=0; k<=3; ++k){
                for(int i=0; i<=2; ++i){
                    LAMBDA[g][j][k][i] = 0.0;
                    GAMMA[g][j][k][i] = 0.0;
                }
            }
        }
    }
  
    LeafGridView leafGridView = grid.leafView();
  
    //FIRSTLOOP OVER ELEMENTS TO ACCESS POINTS IN THE ELEMENT AND IN NEIGHBOR
    // LOOP OVER ELEMENTS (codim 0 entities)
    EIterator endit = leafGridView.template end<codim,ptype>();
    for(EIterator e = leafGridView.template begin<codim,ptype>(); e != endit; ++e)
        {
            const Element el = *e;
            const EGeometry & eGeo = el.geometry();
            GC eCenter = eGeo.center();
            int indexi = mapper.index(el);
      
      
            //INITIALIZE p-STRUCTURE to ZERO to USE THE *= OPERATOR
            for(int u=-1; u<=2; ++u){
                for(int v=-1; v<=2; ++v){
                    for(int w=-1; w<=2; ++w){
                        flagmap[indexi][u][v][w] = 0;
                        for(int d=0; d<=3; ++d){
                            for(int k=0; k<=2; ++k){
                                pmap[indexi][u][v][w][d][k] = 0;
                            }
                        }
                    }
                }
            }
      
      
            //ELEMENT VERTICES
            pmap[indexi][0][0][0][0]=eGeo.corner(0);
            pmap[indexi][1][0][0][0]=eGeo.corner(1);
            pmap[indexi][0][1][0][0]=eGeo.corner(2);
            pmap[indexi][1][1][0][0]=eGeo.corner(3);
            pmap[indexi][0][0][1][0]=eGeo.corner(4);
            pmap[indexi][1][0][1][0]=eGeo.corner(5);
            pmap[indexi][0][1][1][0]=eGeo.corner(6);
            pmap[indexi][1][1][1][0]=eGeo.corner(7);
      
            //NEIGHBORS VERTICES
            IIterator iend = leafGridView.iend(el);
            for(IIterator i = leafGridView.ibegin(el); i != iend; ++i)
                {
                    const Intersection & in = *i;
                    if( in.neighbor() )
                        {
                            int inInsideIdx = in.indexInInside();
                            int inOutsideIdx = in.indexInOutside();
                            const Element outside = in.outside();
                            int indexj = mapper.index(outside);
                            const EGeometry & outsideGeo = outside.geometry();
                            const ILocalGeometry & inLGeoIn = in.geometryInInside();
                            const ILocalGeometry & inLGeoOut = in.geometryInOutside();
	      
                            if( inInsideIdx == 0 )
                                {
                                    for( int k = 0; k < inLGeoIn.corners(); ++k )
                                        {
                                            LC cornerIn = inLGeoIn.corner(k);
                                            int key0 = -1;
                                            int key1 = (int) cornerIn[1];
                                            int key2 = (int) cornerIn[2];

                                            LC cornerOut = inLGeoOut.corner(k);

                                            if(gridPtr.blocks_[indexi]==gridPtr.blocks_[indexj]){
			
                                                if( inOutsideIdx == 0 )
                                                    cornerOut[0] = 1;
                                                else if( inOutsideIdx == 1 )
                                                    cornerOut[0] = 0;
                                                else if( inOutsideIdx == 2 )
                                                    cornerOut[1] = 1;
                                                else if( inOutsideIdx == 3 )
                                                    cornerOut[1] = 0;
                                                else if( inOutsideIdx == 4 )
                                                    cornerOut[2] = 1;
                                                else if( inOutsideIdx == 5 )
                                                    cornerOut[2] = 0;

                                                pmap[indexi][key0][key1][key2][0] = outsideGeo.global(cornerOut);
                                            }
                                            else
                                                pmap[indexi][key0][key1][key2][0] = computePointOverBoundary(pmap[indexi][key0+1][key1][key2][0],pmap[indexi][key0+2][key1][key2][0]);

                                            flagmap[indexi][key0][key1][key2] = 1;

                                        }//end of for over intersection corners
                                }//end of if inside index is 0

                            else if( inInsideIdx == 1 )
                                {
                                    for( int k = 0; k < inLGeoIn.corners(); ++k )
                                        {
                                            LC cornerIn = inLGeoIn.corner(k);
                                            int key0 = 2;
                                            int key1 = (int) cornerIn[1];
                                            int key2 = (int) cornerIn[2];

                                            LC cornerOut = inLGeoOut.corner(k);

                                            if(gridPtr.blocks_[indexi]==gridPtr.blocks_[indexj]){

                                                if( inOutsideIdx == 0 )
                                                    cornerOut[0] = 1;
                                                else if( inOutsideIdx == 1 )
                                                    cornerOut[0] = 0;
                                                else if( inOutsideIdx == 2 )
                                                    cornerOut[1] = 1;
                                                else if( inOutsideIdx == 3 )
                                                    cornerOut[1] = 0;
                                                else if( inOutsideIdx == 4 )
                                                    cornerOut[2] = 1;
                                                else if( inOutsideIdx == 5 )
                                                    cornerOut[2] = 0;

                                                pmap[indexi][key0][key1][key2][0] = outsideGeo.global(cornerOut);
                                            }
                                            else
                                                pmap[indexi][key0][key1][key2][0] = computePointOverBoundary(pmap[indexi][key0-1][key1][key2][0],pmap[indexi][key0-2][key1][key2][0]);

                                            flagmap[indexi][key0][key1][key2] = 1;

                                        }//end of for over intersection corners
                                }//end of if inside index is 1

                            else if( inInsideIdx == 2 )
                                {
                                    for( int k = 0; k < inLGeoIn.corners(); ++k )
                                        {
                                            LC cornerIn = inLGeoIn.corner(k);
                                            int key0 = (int) cornerIn[0];
                                            int key1 = -1;
                                            int key2 = (int) cornerIn[2];

                                            LC cornerOut = inLGeoOut.corner(k);

                                            // if( inOutsideIdx == 0 )
                                            // 	cornerOut[0] = 1;
                                            // if( inOutsideIdx == 1 )
                                            // 	cornerOut[0] = 0;
                                            // if( inOutsideIdx == 2 )
                                            // 	cornerOut[1] = 1;
                                            if(gridPtr.blocks_[indexi]==gridPtr.blocks_[indexj]){

                                                if( inOutsideIdx == 0 )
                                                    cornerOut[0] = 1;
                                                else if( inOutsideIdx == 1 )
                                                    cornerOut[0] = 0;
                                                else if( inOutsideIdx == 2 )
                                                    cornerOut[1] = 1;
                                                else if( inOutsideIdx == 3 )
                                                    cornerOut[1] = 0;
                                                else if( inOutsideIdx == 4 )
                                                    cornerOut[2] = 1;
                                                else if( inOutsideIdx == 5 )
                                                    cornerOut[2] = 0;

                                                pmap[indexi][key0][key1][key2][0] = outsideGeo.global(cornerOut);
                                            }
                                            else
                                                pmap[indexi][key0][key1][key2][0] = computePointOverBoundary(pmap[indexi][key0][key1+1][key2][0],pmap[indexi][key0][key1+2][key2][0]);

                                            flagmap[indexi][key0][key1][key2] = 1;

                                        }//end of for over intersection corners


                                }//end of if inside index is 2

                            else if( inInsideIdx == 3 )
                                {
                                    for( int k = 0; k < inLGeoIn.corners(); ++k )
                                        {
                                            LC cornerIn = inLGeoIn.corner(k);
                                            int key0 = (int) cornerIn[0];
                                            int key1 = 2;
                                            int key2 = (int) cornerIn[2];

                                            LC cornerOut = inLGeoOut.corner(k);

                                            if(gridPtr.blocks_[indexi]==gridPtr.blocks_[indexj]){

                                                if( inOutsideIdx == 0 )
                                                    cornerOut[0] = 1;
                                                else if( inOutsideIdx == 1 )
                                                    cornerOut[0] = 0;
                                                else if( inOutsideIdx == 2 )
                                                    cornerOut[1] = 1;
                                                else if( inOutsideIdx == 3 )
                                                    cornerOut[1] = 0;
                                                else if( inOutsideIdx == 4 )
                                                    cornerOut[2] = 1;
                                                else if( inOutsideIdx == 5 )
                                                    cornerOut[2] = 0;

                                                pmap[indexi][key0][key1][key2][0] = outsideGeo.global(cornerOut);
                                            }
                                            else
                                                pmap[indexi][key0][key1][key2][0] = computePointOverBoundary(pmap[indexi][key0][key1-1][key2][0],pmap[indexi][key0][key1-2][key2][0]);
		      
                                            flagmap[indexi][key0][key1][key2] = 1;

                                        }//end of for over intersection corners

                                }//end of if inside index is 3

                            else if( inInsideIdx == 4 )
                                {
                                    for( int k = 0; k < inLGeoIn.corners(); ++k )
                                        {
                                            LC cornerIn = inLGeoIn.corner(k);
                                            int key0 = (int) cornerIn[0];
                                            int key1 = (int) cornerIn[1];
                                            int key2 = -1;

                                            LC cornerOut = inLGeoOut.corner(k);

                                            if(gridPtr.blocks_[indexi]==gridPtr.blocks_[indexj]){

                                                if( inOutsideIdx == 0 )
                                                    cornerOut[0] = 1;
                                                else if( inOutsideIdx == 1 )
                                                    cornerOut[0] = 0;
                                                else if( inOutsideIdx == 2 )
                                                    cornerOut[1] = 1;
                                                else if( inOutsideIdx == 3 )
                                                    cornerOut[1] = 0;
                                                else if( inOutsideIdx == 4 )
                                                    cornerOut[2] = 1;
                                                else if( inOutsideIdx == 5 )
                                                    cornerOut[2] = 0;

                                                pmap[indexi][key0][key1][key2][0] = outsideGeo.global(cornerOut);
                                            }
                                            else
                                                pmap[indexi][key0][key1][key2][0] = computePointOverBoundary(pmap[indexi][key0][key1][key2+1][0],pmap[indexi][key0][key1][key2+2][0]);

                                            flagmap[indexi][key0][key1][key2] = 1;

                                        }//end of for over intersection corners

                                }//end of if inside index is 4

                            else if( inInsideIdx == 5 )
                                {
                                    for( int k = 0; k < inLGeoIn.corners(); ++k )
                                        {
                                            LC cornerIn = inLGeoIn.corner(k);
                                            int key0 = (int) cornerIn[0];
                                            int key1 = (int) cornerIn[1];
                                            int key2 = 2;

                                            LC cornerOut = inLGeoOut.corner(k);

                                            if(gridPtr.blocks_[indexi]==gridPtr.blocks_[indexj]){

                                                if( inOutsideIdx == 0 )
                                                    cornerOut[0] = 1;
                                                else if( inOutsideIdx == 1 )
                                                    cornerOut[0] = 0;
                                                else if( inOutsideIdx == 2 )
                                                    cornerOut[1] = 1;
                                                else if( inOutsideIdx == 3 )
                                                    cornerOut[1] = 0;
                                                else if( inOutsideIdx == 4 )
                                                    cornerOut[2] = 1;
                                                else if( inOutsideIdx == 5 )
                                                    cornerOut[2] = 0;

                                                pmap[indexi][key0][key1][key2][0] = outsideGeo.global(cornerOut);
                                            }
                                            else
                                                pmap[indexi][key0][key1][key2][0] = computePointOverBoundary(pmap[indexi][key0][key1][key2-1][0],pmap[indexi][key0][key1][key2-2][0]);

                                            flagmap[indexi][key0][key1][key2] = 1;

                                        }//end of for over intersection corners

                                }//end of if inside index is 5


	      
                        }//end of if in is neighbor
                    // else{
                    //   //DEBUG-----------------------------------------------------------------------------------------------------------------------------------------------------
                    //   int inInsideIdx = in.indexInInside();
                    //   const ILocalGeometry & inLGeoIn = in.geometryInInside();
                    //   if( inInsideIdx == 0 )
                    //     {
                    // 	for( int k = 0; k < inLGeoIn.corners(); ++k )
                    // 	  {
                    // 	    LC cornerIn = inLGeoIn.corner(k);
                    // 	    int key0 = -1;
                    // 	    int key1 = (int) cornerIn[1];
                    // 	    int key2 = (int) cornerIn[2];
		      
                    // 	    pmap[indexi][key0][key1][key2][0] = computePointOverBoundary(pmap[indexi][key0+1][key1][key2][0],pmap[indexi][key0+2][key1][key2][0]);
                    // 	    flagmap[indexi][key0][key1][key2] = 1;

                    // 	  }//end of for over intersection corners
                    //     }//end of if inside index is 0

                    //   if( inInsideIdx == 1 )
                    //     {
                    // 	for( int k = 0; k < inLGeoIn.corners(); ++k )
                    // 	  {
                    // 	    LC cornerIn = inLGeoIn.corner(k);
                    // 	    int key0 = 2;
                    // 	    int key1 = (int) cornerIn[1];
                    // 	    int key2 = (int) cornerIn[2];

                    // 	    pmap[indexi][key0][key1][key2][0] = computePointOverBoundary(pmap[indexi][key0-1][key1][key2][0],pmap[indexi][key0-2][key1][key2][0]);
                    // 	    flagmap[indexi][key0][key1][key2] = 1;

                    // 	  }//end of for over intersection corners
                    //     }//end of if inside index is 1

                    //   if( inInsideIdx == 2 )
                    //     {
                    // 	for( int k = 0; k < inLGeoIn.corners(); ++k )
                    // 	  {
                    // 	    LC cornerIn = inLGeoIn.corner(k);
                    // 	    int key0 = (int) cornerIn[0];
                    // 	    int key1 = -1;
                    // 	    int key2 = (int) cornerIn[2];

                    // 	    pmap[indexi][key0][key1][key2][0] = computePointOverBoundary(pmap[indexi][key0][key1+1][key2][0],pmap[indexi][key0][key1+2][key2][0]);
                    // 	    flagmap[indexi][key0][key1][key2] = 1;

                    // 	  }//end of for over intersection corners


                    //     }//end of if inside index is 2

                    //   if( inInsideIdx == 3 )
                    //     {
                    // 	for( int k = 0; k < inLGeoIn.corners(); ++k )
                    // 	  {
                    // 	    LC cornerIn = inLGeoIn.corner(k);
                    // 	    int key0 = (int) cornerIn[0];
                    // 	    int key1 = 2;
                    // 	    int key2 = (int) cornerIn[2];

                    // 	    pmap[indexi][key0][key1][key2][0] = computePointOverBoundary(pmap[indexi][key0][key1-1][key2][0],pmap[indexi][key0][key1-2][key2][0]);
                    // 	    flagmap[indexi][key0][key1][key2] = 1;

                    // 	  }//end of for over intersection corners

                    //     }//end of if inside index is 3

                    //   if( inInsideIdx == 4 )
                    //     {
                    // 	for( int k = 0; k < inLGeoIn.corners(); ++k )
                    // 	  {
                    // 	    LC cornerIn = inLGeoIn.corner(k);
                    // 	    int key0 = (int) cornerIn[0];
                    // 	    int key1 = (int) cornerIn[1];
                    // 	    int key2 = -1;

                    // 	    pmap[indexi][key0][key1][key2][0] = computePointOverBoundary(pmap[indexi][key0][key1][key2+1][0],pmap[indexi][key0][key1][key2+2][0]);
                    // 	    flagmap[indexi][key0][key1][key2] = 1;

                    // 	  }//end of for over intersection corners

                    //     }//end of if inside index is 4

                    //   if( inInsideIdx == 5 )
                    //     {
                    // 	for( int k = 0; k < inLGeoIn.corners(); ++k )
                    // 	  {
                    // 	    LC cornerIn = inLGeoIn.corner(k);
                    // 	    int key0 = (int) cornerIn[0];
                    // 	    int key1 = (int) cornerIn[1];
                    // 	    int key2 = 2;


                    // 	    pmap[indexi][key0][key1][key2][0] = computePointOverBoundary(pmap[indexi][key0][key1][key2-1][0],pmap[indexi][key0][key1][key2-2][0]);
                    // 	    flagmap[indexi][key0][key1][key2] = 1;

                    // 	  }//end of for over intersection corners

                    //     }//end of if inside index is 5

                    //   //DEBUG-----------------------------------------------------------------------------------------------------------------------------------------------------
                    // }////end of else in is neighbor
                }//end loop over intersections
        }//end loop over elements
  
    PMAPExchange< M, std::vector< PmapType > > pmapDataHandle (mapper,pmap);
    grid.template communicate< PMAPExchange< M, std::vector< PmapType> > >(pmapDataHandle,Dune::InteriorBorder_All_Interface,Dune::ForwardCommunication);
    FLAGMAPExchange< M, std::vector < FlagmapType > > flagmapDataHandle (mapper,flagmap);
    grid.template communicate< FLAGMAPExchange< M, std::vector< FlagmapType > > >(flagmapDataHandle,Dune::InteriorBorder_All_Interface, Dune::ForwardCommunication);
  
    Dune::PersistentContainer<G,int> changedVertex(grid,dim);

    //SECOND LOOP OVER ELEMENTS TO ACCESS POINTS ON THE BOUNDARY EDGE
    for(EIterator e = leafGridView.template begin<codim,ptype>(); e != endit; ++e)
        {
            const Element el = *e;
            const EGeometry & eGeo = el.geometry();
            GC eCenter = eGeo.center();
            int indexi = mapper.index(el);
            int indexi3 = indexi;
      
            // for(int g=0; g<=3; ++g){
            //  	for(int j=0; j<=3; ++j){
            //  	  for(int k=0; k<=3; ++k){
            //  	    for(int i=0; i<=2; ++i){
            //  	      LAMBDA[g][j][k][i] = 0.0;
            //  	      GAMMA[g][j][k][i] = 0.0;
            //  	    }
            //  	  }
            //  	}
            // }


            ///////////////////////////////////////
            IIterator iend = leafGridView.iend(el);
            for(IIterator ii = leafGridView.ibegin(el); ii != iend; ++ii)
                {
                    const Intersection & in = *ii;
                    int inInsideIdx = in.indexInInside();
	  
                    if( in.neighbor() )
                        {

                            const Element outside = in.outside();
                            const EGeometry & outsideGeo = outside.geometry();
                            int indexj = mapper.index( outside );
	      
                            int inOutsideIdx = in.indexInOutside();
                            const ILocalGeometry & inLGeoIn = in.geometryInInside();
                            const ILocalGeometry & inLGeoOut = in.geometryInOutside();
	      
                            for(int kk = 0; kk < 2; ++kk)
                                {
                                    for(int mm = 0; mm < 2; ++mm)
                                        {
                                            int i = mm + kk*mm;
                                            int j = kk + 2*mm + 1 - kk*mm;
		      
                                            LC c0In = inLGeoIn.corner(i);
                                            LC c1In = inLGeoIn.corner(j);
                                            LC cornerDiffIn = c0In;
                                            cornerDiffIn -= c1In;
		      
                                            LC c0Out = inLGeoOut.corner(i);
                                            LC c1Out = inLGeoOut.corner(j);
                                            LC cornerDiffOut = c0Out;
                                            cornerDiffOut -= c1Out;
		      
                                            LC keysPIn = c0In;
                                            keysPIn += cornerDiffIn;
                                            int key0PIn = (int)keysPIn[0]; int key1PIn = (int)keysPIn[1]; int key2PIn = (int)keysPIn[2];
                                            LC keysMIn = c1In;
                                            keysMIn -= cornerDiffIn;
                                            int key0MIn = (int)keysMIn[0]; int key1MIn = (int)keysMIn[1]; int key2MIn = (int)keysMIn[2];
		      
                                            LC keysPOut = c0Out;
                                            keysPOut += cornerDiffOut;
                                            int key0POut = (int)keysPOut[0]; int key1POut = (int)keysPOut[1]; int key2POut = (int)keysPOut[2];
                                            LC keysMOut = c1Out;
                                            keysMOut -= cornerDiffOut;
                                            int key0MOut = (int)keysMOut[0]; int key1MOut = (int)keysMOut[1]; int key2MOut = (int)keysMOut[2];
		      
                                            if( flagmap[indexi][ key0PIn ][ key1PIn ][ key2PIn ] == 0 && flagmap[indexj][ key0POut ][ key1POut ][ key2POut ] == 1)
                                                {
                                                    pmap[indexi][key0PIn][key1PIn][key2PIn][0] = pmap[indexj][key0POut][key1POut][key2POut][0];
                                                    flagmap[indexi][key0PIn][key1PIn][key2PIn] = 1;
                                                }
		      
                                            if( flagmap[indexi][ key0MIn ][ key1MIn ][ key2MIn ] == 0 && flagmap[indexj][ key0MOut ][ key1MOut ][ key2MOut ] == 1)
                                                {
                                                    pmap[indexi][key0MIn][key1MIn][key2MIn][0] = pmap[indexj][key0MOut][key1MOut][key2MOut][0];
                                                    flagmap[indexi][key0MIn][key1MIn][key2MIn] = 1;
                                                }
		      
                                            int ciccio = 0;
                                        }//end for mm
                                }//end for kk
                        }//end of if neighbor
                }//end of loop over intersections

            //OUT-OF-BOUNDARIES POINTS
            for(int k = -1; k < 3; k+=3)
                {
                    for(int i = 0; i < 2; ++i)
                        {
                            for(int j = 0; j < 2; ++j)
                                {
                                    int k1 = (k + 1)/3;
                                    int k2 = (-k + 2)/3;
                                    if(flagmap[indexi][k][i][j] == 0)
                                        {
                                            pmap[indexi][k][i][j][0] = computePointOverBoundary(pmap[indexi][k1][i][j][0],pmap[indexi][k2][i][j][0]);
                                            //		      flagmap[indexi][k][i][j] = 1;
                                        }
                                    if(flagmap[indexi][i][k][j] == 0)
                                        {
                                            pmap[indexi][i][k][j][0] = computePointOverBoundary(pmap[indexi][i][k1][j][0],pmap[indexi][i][k2][j][0]);
                                            //		      flagmap[indexi][i][k][j] =1;
                                        }
                                    if(flagmap[indexi][i][j][k] == 0)
                                        {
                                            pmap[indexi][i][j][k][0] = computePointOverBoundary(pmap[indexi][i][j][k1][0],pmap[indexi][i][j][k2][0]);
                                            //		      flagmap[indexi][i][j][k] = 1;
                                        }
                                }
                        }
                }

            //-------------//
            // DERIVATIVES //
            //-------------//
            // along u - pmap[u][v][w][1] with u,v,w \in {0,1}
            // along v - pmap[u][v][w][2] with u,v,w \in {0,1}
            // along w - pmap[u][v][w][3] with u,v,w \in {0,1}
            for(int u=0; u<=1; ++u){
                for(int v=0; v<=1; ++v){
                    for(int w=0; w<=1; ++w){
                        pmap[indexi][u][v][w][1] = computeDerivatives(pmap[indexi][u+1][v][w][0],pmap[indexi][u-1][v][w][0]);
                        pmap[indexi][u][v][w][2] = computeDerivatives(pmap[indexi][u][v+1][w][0],pmap[indexi][u][v-1][w][0]);
                        pmap[indexi][u][v][w][3] = computeDerivatives(pmap[indexi][u][v][w+1][0],pmap[indexi][u][v][w-1][0]);
                    }
                }
            }

        }

    grid.template communicate< PMAPExchange< M, std::vector< PmapType> > >(pmapDataHandle,Dune::InteriorBorder_All_Interface,Dune::ForwardCommunication);
    grid.template communicate< FLAGMAPExchange< M, std::vector< FlagmapType > > >(flagmapDataHandle,Dune::InteriorBorder_All_Interface, Dune::ForwardCommunication);

    //controllo derivate
    for(EIterator e = leafGridView.template begin<codim,ptype>(); e != endit; ++e)
        {
            const Element el = *e;
            const EGeometry & eGeo = el.geometry();
            GC eCenter = eGeo.center();
            int indexi = mapper.index(el);
            int indexi4 = indexi;

            IIterator iend = leafGridView.iend(el);
            for(IIterator ii = leafGridView.ibegin(el); ii != iend; ++ii)
                {
                    const Intersection & in = *ii;
                    if( in.neighbor() )
                        {
                            int inInsideIdx = in.indexInInside();
                            int inOutsideIdx = in.indexInOutside();
                            const Element outside = in.outside();
                            int indexj = mapper.index(outside);
                            const EGeometry & outsideGeo = outside.geometry();
                            GC outsideCenter = outsideGeo.center();
                            const ILocalGeometry & inLGeoIn = in.geometryInInside();
                            const ILocalGeometry & inLGeoOut = in.geometryInOutside();
	      
                            for(int kk = 0; kk < 2; ++kk)
                                {
                                    for(int mm = 0; mm < 2; ++mm)
                                        {
                                            int i = mm + kk*mm;
                                            int j = kk + 2*mm + 1 - kk*mm;

                                            int derivativeIdxIn = 4;
                                            int signIn = 0;
                                            int derivativeIdxOut = 4;
                                            int signOut = 0;


                                            LC c0In = inLGeoIn.corner(i);
                                            LC c1In = inLGeoIn.corner(j);
                                            LC cornerDiffIn = c0In;
                                            cornerDiffIn -= c1In;
                                            int c0key0In = (int) c0In[0];
                                            int c0key1In = (int) c0In[1];
                                            int c0key2In = (int) c0In[2];
                                            int c1key0In = (int) c1In[0];
                                            int c1key1In = (int) c1In[1];
                                            int c1key2In = (int) c1In[2];
                                            for(int k = 0; k < dim; ++k)
                                                {
                                                    if(cornerDiffIn[k] > 0.5 || cornerDiffIn[k] < -0.5){
                                                        derivativeIdxIn = k+1;
                                                        signIn = (int)cornerDiffIn[k];
                                                        break;
                                                    }
                                                }
		      
                                            LC c0Out = inLGeoOut.corner(i);
                                            LC c1Out = inLGeoOut.corner(j);
                                            LC cornerDiffOut = c0Out;
                                            cornerDiffOut -= c1Out;
                                            int c0key0Out = (int) c0Out[0];
                                            int c0key1Out = (int) c0Out[1];
                                            int c0key2Out = (int) c0Out[2];
                                            int c1key0Out = (int) c1Out[0];
                                            int c1key1Out = (int) c1Out[1];
                                            int c1key2Out = (int) c1Out[2];
                                            for(int k = 0; k < dim; ++k)
                                                {
                                                    if(cornerDiffOut[k] > 0.5 || cornerDiffOut[k] < -0.5){
                                                        derivativeIdxOut = k+1;
                                                        signOut = (int)cornerDiffOut[k];
                                                        break;
                                                    }
                                                }
		      
                                            assert(derivativeIdxIn < 4);
                                            assert(derivativeIdxOut < 4);

                                            // if( inInsideIdx == 0 || inInsideIdx == 1)
                                            //   derivativeIdxIn = 1;
                                            // else if( inInsideIdx == 2 || inInsideIdx == 3) 
                                            //   derivativeIdxIn = 2;
                                            // else if( inInsideIdx == 4 || inInsideIdx == 5 )
                                            //   derivativeIdxIn = 3;
                                            // assert(derivativeIdxIn < 4);
                                            // if( inOutsideIdx == 0 || inOutsideIdx == 1)
                                            //   derivativeIdxOut = 1;
                                            // else if( inOutsideIdx == 2 || inOutsideIdx == 3) 
                                            //   derivativeIdxOut = 2;
                                            // else if( inOutsideIdx == 4 || inOutsideIdx == 5 )
                                            //   derivativeIdxOut = 3;
                                            // assert(derivativeIdxOut < 4);

                                            if(gridPtr.blocks_[indexi] == gridPtr.blocks_[indexj]){
			
                                                if(!derivativeModuleEqual(pmap[indexi][c0key0In][c0key1In][c0key2In][derivativeIdxIn],pmap[indexj][c0key0Out][c0key1Out][c0key2Out][derivativeIdxOut])){

                                                    bool derivativeCheck0 = false;
                                                    for(int k = 0; k < dim; ++k)
                                                        {
                                                            if( fabs( pmap[indexi][c0key0In][c0key1In][c0key2In][derivativeIdxIn][k] - pmap[indexj][c0key0Out][c0key1Out][c0key2Out][derivativeIdxOut][k] ) > 1e-6 )
                                                                {
                                                                    derivativeCheck0 = true;
                                                                    break;
                                                                }
                                                        }
			  
			  
                                                    if(derivativeCheck0)
                                                        {
                                                            GC newDerivative = eGeo.global( c0In ) - eGeo.global( c1In );
                                                            newDerivative *= (double)signIn;
                                                            pmap[indexi][c0key0In][c0key1In][c0key2In][derivativeIdxIn] = newDerivative;
                                                        }
			  
                                                }
			
                                                if(!derivativeModuleEqual(pmap[indexi][c1key0In][c1key1In][c1key2In][derivativeIdxIn],pmap[indexj][c1key0Out][c1key1Out][c1key2Out][derivativeIdxOut])){
                                                    bool derivativeCheck1 = false;
                                                    for(int k = 0; k < dim; ++k)
                                                        {
                                                            if( fabs( pmap[indexi][c1key0In][c1key1In][c1key2In][derivativeIdxIn][k] - pmap[indexj][c1key0Out][c1key1Out][c1key2Out][derivativeIdxOut][k] ) > 1e-6 )
                                                                {
                                                                    derivativeCheck1 = true;
                                                                    break;
                                                                }
                                                        }
			  
                                                    if(derivativeCheck1)
                                                        {
                                                            GC newDerivative = eGeo.global( c1In ) - eGeo.global( c0In );
                                                            newDerivative *= (double)( signIn * (-1) );
                                                            pmap[indexi][c1key0In][c1key1In][c1key2In][derivativeIdxIn] = newDerivative;
                                                        }
                                                }
                                            }//end if blocks equal
		      
                                        }//end of loops over intersection corners(segments)
                                }//end of loops over intersection corners

                        }//end of if neighbor
                }//end of intersection loop

        }//end of second loop over elements

    //third loop over elements
    for(EIterator e = leafGridView.template begin<codim,ptype>(); e != endit; ++e)
        {
            const Element el = *e;
            const EGeometry & eGeo = el.geometry();
            GC eCenter = eGeo.center();
            int indexi = mapper.index(el);

            // IIterator iend = leafGridView.iend(el);
            // for(IIterator ii = leafGridView.ibegin(el); ii != iend; ++ii)
            // 	{
            // 	  const Intersection & in = *ii;
            // 	  if( in.neighbor() )
            // 	    {
            // 	      int inInsideIdx = in.indexInInside();
            // 	      int inOutsideIdx = in.indexInOutside();
            // 	      EPointer outsidePointer = in.outside();
            // 	      Element & outside = *outsidePointer;
            // 	      int indexj = mapper.map(outside);
            // 	      //	      const EGeometry & outsideGeo = outside.geometry();
            // 	      //GC outsideCenter = outsideGeo.center();
            // 	      const ILocalGeometry & inLGeoIn = in.geometryInInside();
            // 	      const ILocalGeometry & inLGeoOut = in.geometryInOutside();
      
            // 	      for(int k = 0; k < inLGeoIn.corners(); ++k)
            // 		{
		  
            // 		  int normalDerivativeIdxIn = 4;
            // 		  int normalDerivativeSignIn = 0;
            // 		  int normalDerivativeIdxOut = 4;
            // 		  int normalDerivativeSignOut = 0;
		      
            // 		  if( inInsideIdx%2 == 0 )
            // 		    normalDerivativeSignIn = 1;
            // 		  else
            // 		    normalDerivativeSignIn = -1;

            // 		  if( inInsideIdx == 0 || inInsideIdx == 1 )
            // 		    normalDerivativeIdxIn = 1;
            // 		  else if ( inInsideIdx == 2 || inInsideIdx == 3 )
            // 		    normalDerivativeIdxIn = 2;
            // 		  else if ( inInsideIdx == 4 || inInsideIdx == 5 )
            // 		    normalDerivativeIdxIn = 3;
		  
            // 		  // if( inOutsideIdx%2 == 0 )
            // 		  //   normalDerivativeSignOut = 1;
            // 		  // else
            // 		  //   normalDerivativeSignOut = -1;
            // 		  if( inOutsideIdx == 0 || inOutsideIdx == 1 )
            // 		    normalDerivativeIdxOut = 1;
            // 		  else if ( inOutsideIdx == 2 || inOutsideIdx == 3 )
            // 		    normalDerivativeIdxOut = 2;
            // 		  else if ( inOutsideIdx == 4 || inOutsideIdx == 5 )
            // 		    normalDerivativeIdxOut = 3;
		  
            // 		  LC c0In = inLGeoIn.corner(k);
            // 		  LC c0Out = inLGeoOut.corner(k);
            // 		  int c0key0In = (int) c0In[0];
            // 		  int c0key1In = (int) c0In[1];
            // 		  int c0key2In = (int) c0In[2];
            // 		  int c0key0Out = (int) c0Out[0];
            // 		  int c0key1Out = (int) c0Out[1];
            // 		  int c0key2Out = (int) c0Out[2];
		  
            // 		  bool derivativeCheck0 = false;
            // 		  for(int kk = 0; kk < dim; ++kk)
            // 		    {
            // 		      if( fabs( pmap[indexi][c0key0In][c0key1In][c0key2In][normalDerivativeIdxIn][kk] - pmap[indexj][c0key0Out][c0key1Out][c0key2Out][normalDerivativeIdxOut][kk] ) > 1e-6 )
            // 			{
            // 			  derivativeCheck0 = true;
            // 			  break;
            // 			}
            // 		    }
		  
		  
            // 		  if(derivativeCheck0)
            // 		    {
            // 		      LC c1In = c0In; c1In[ normalDerivativeIdxIn - 1 ] += normalDerivativeSignIn;
            // 		      GC newDerivative = eGeo.global( c0In ) - eGeo.global( c1In );
            // 		      newDerivative *= (double)normalDerivativeSignIn * (-1.0);
            // 		      pmap[indexi][c0key0In][c0key1In][c0key2In][normalDerivativeIdxIn] = newDerivative;
            // 		      // newDerivative = outsideGeo.global( c0Out ) - outsideGeo.global( c1Out );
            // 		      // newDerivatives *= (double)signOut;
            // 		      //pmap[indexj][c0key0Out][c0key1Out][c0key2Out][derivativeIdxOut] = newDerivative;
            // 		    }
		  

            // 		}//end of intersection corners


            // 	    }
            // 	}



            for(int g=0; g<=3; ++g){
                for(int j=0; j<=3; ++j){
                    for(int k=0; k<=3; ++k){
                        for(int i=0; i<=2; ++i){
                            LAMBDA[g][j][k][i] = 0.0;
                            GAMMA[g][j][k][i] = 0.0;
                        }
                    }
                }
            }
            //-------------------------------------------------//
            // COMPUTE LAMBDA COEFFICIENTS - 2nd INTERPOLATION //
            //-------------------------------------------------//
            for(int g = 0; g <= 3; ++g){
                for(int j = 0; j <= 3; ++j){
                    for(int k = 0; k <= 3; ++k){
                        for(int i = 0; i <= 2; ++i){
	      
                            for(int b = 0; b <= 1; ++b){
                                for(int l = 0; l <= 1; ++l){
                                    for(int m = 0; m <= 1; ++m){
		    
                                        for(int p = 0; p <= 1; ++p){
                                            for(int q = 0; q <= 1; ++q){
                                                if(p==0){
                                                    buffer = (pmap[indexi][b][q][l][0][i] * alpha[p][q][j] * alpha[0][l][k] * alpha[0][b][g] + pmap[indexi][b][m][q][0][i] * alpha[p][q][k] * alpha[0][m][j] * alpha[0][b][g]) * kronecker(l,m);
                                                }
                                                else if(p==1){
                                                    buffer = (pmap[indexi][b][q][l][2][i] * alpha[p][q][j] * alpha[0][l][k] * alpha[0][b][g] + pmap[indexi][b][m][q][3][i] * alpha[p][q][k] * alpha[0][m][j] * alpha[0][b][g]) * kronecker(l,m);
                                                }
                                                LAMBDA[g][j][k][i] += buffer;
                                            } // closing q-loop
                                        } // closing p-loop
                                        buffer = -pmap[indexi][b][l][m][0][i] * alpha[0][l][j] * alpha[0][m][k] * alpha[0][b][g];
                                        LAMBDA[g][j][k][i] += buffer;
		    
                                        for(int p = 0; p <= 1; ++p){
                                            for(int q = 0; q <= 1; ++q){
                                                if(p==0){
                                                    buffer = (pmap[indexi][q][b][l][0][i] * alpha[p][q][g] * alpha[0][l][k] * alpha[0][b][j] + pmap[indexi][m][b][q][0][i] * alpha[p][q][k] * alpha[0][m][g] * alpha[0][b][j]) * kronecker(l,m);
                                                }
                                                else if(p==1){
                                                    buffer = (pmap[indexi][q][b][l][1][i] * alpha[p][q][g] * alpha[0][l][k] * alpha[0][b][j] + pmap[indexi][m][b][q][3][i] * alpha[p][q][k] * alpha[0][m][g] * alpha[0][b][j]) * kronecker(l,m);
                                                }
                                                LAMBDA[g][j][k][i] += buffer;
                                            } // closing q-loop
                                        } // closing p-loop
                                        buffer = -pmap[indexi][l][b][m][0][i] * alpha[0][l][g] * alpha[0][m][k] * alpha[0][b][j];
                                        LAMBDA[g][j][k][i] += buffer;
		    
                                    } // closing m-loop
                                } // closing l-loop
                            } // closing b-loop
	      
                            for(int s = 0; s <= 3; ++s){
                                for(int b = 0; b <= 1; ++b){
                                    for(int a = 0; a <= 1; ++a){
                                        for(int l = 0; l <= 1; ++l){
                                            for(int m = 0; m <= 1; ++m){
			
                                                for(int p = 0; p <= 1; ++p){
                                                    for(int q = 0; q <= 1; ++q){
                                                        if(p==0){
                                                            buffer = (-pmap[indexi][a][q][l][0][i] * alpha[p][q][s] * alpha[0][l][k] * alpha[0][a][g] * alpha[0][b][j] - pmap[indexi][a][m][q][0][i] * alpha[p][q][k] * alpha[0][m][s] * alpha[0][a][g] * alpha[0][b][j]) * kronecker(l,m) * Pow(b,s);
                                                        }
                                                        else if(p==1){
                                                            buffer = (-pmap[indexi][a][q][l][2][i] * alpha[p][q][s] * alpha[0][l][k] * alpha[0][a][g] * alpha[0][b][j] - pmap[indexi][a][m][q][3][i] * alpha[p][q][k] * alpha[0][m][s] * alpha[0][a][g] * alpha[0][b][j]) * kronecker(l,m) * Pow(b,s);
                                                        }
                                                        LAMBDA[g][j][k][i] += buffer;
                                                    } // closing q-loop
                                                } // closing p-loop
                                                buffer = pmap[indexi][a][l][m][0][i] * alpha[0][l][s] * alpha[0][m][k] * alpha[0][a][g] * alpha[0][b][j] * Pow(b,s);
                                                LAMBDA[g][j][k][i] += buffer;
			
                                            } // closing m-loop
                                        } // closing l-loop
                                    } // closing a-loop
                                } // closing b-loop
                            } // closing s-loop
	      
                        } // closing i-loop
                    } // closing k-loop
                } // closing j-loop
            }; // closing g-loop

            //------------------------------------------------------------------------------------//
            // COMPUTE GAMMA COEFFICIENTS - 3rd AND FINAL  INTERPOLATION //
            //------------------------------------------------------------------------------------//
      
            for(int g = 0; g <= 3; ++g){
                for(int j = 0; j <= 3; ++j){
                    for(int k = 0; k <= 3; ++k){
                        for(int i = 0; i <= 2; ++i){
	      
                            GAMMA[g][j][k][i] = LAMBDA[g][j][k][i];
	      
                            for(int c = 0; c <= 1; ++c){
                                for(int l = 0; l <= 1; ++l){
                                    for(int m = 0; m <= 1; ++m){
		    
                                        for(int p = 0; p <= 1; ++p){
                                            for(int q = 0; q <= 1; ++q){
                                                if(p==0){
                                                    buffer = ( pmap[indexi][q][l][c][0][i] * alpha[p][q][g] * alpha[0][l][j] * alpha[0][c][k] + pmap[indexi][m][q][c][0][i] * alpha[p][q][j] * alpha[0][m][g] * alpha[0][c][k]) * kronecker(l,m);
                                                    GAMMA[g][j][k][i] += buffer;
                                                }
                                                else if(p==1){
                                                    buffer = ( pmap[indexi][q][l][c][1][i] * alpha[p][q][g] * alpha[0][l][j] * alpha[0][c][k] + pmap[indexi][m][q][c][2][i] * alpha[p][q][j] * alpha[0][m][g] * alpha[0][c][k]) * kronecker(l,m);
                                                    GAMMA[g][j][k][i] += buffer;
                                                }
                                            }
                                        }
                                        buffer = -pmap[indexi][l][m][c][0][i] * alpha[0][l][g] * alpha[0][m][j] * alpha[0][c][k];
                                        GAMMA[g][j][k][i] += buffer;
                                    }
                                }
                            }
	      
                            for(int h = 0; h <= 3; ++h){
                                for(int c = 0; c <= 1; ++c){
                                    buffer = -alpha[0][c][k] * LAMBDA[g][j][h][i] * Pow(c,h);
                                    GAMMA[g][j][k][i] += buffer;
                                }
                            }
	      
                        }//end of i loop
                    }//end of k loop
                }//end of j loop
            }//end of g loop
      
            //-------------------------------------------------------------------------------//
            // BUILDING ELEMENT PARAMETRIZATION AND STORING IT IN PARAMETRIZATIONS ATTRIBUTE //
            //-------------------------------------------------------------------------------//
            MyParametrization<RangeVector,RangeVector> testparametrization(GAMMA);
            container[el] = testparametrization;// se il container è un persistentcontainer
      
        }//end of second loop over elements
  
  
  
};//end of function computeparametrizationsUNSTRUCTURED


//****************************//
// PMAP e FLAGMAP DATAHANDLES //
//****************************//

template <class M, class V>
class PMAPExchange : public Dune::CommDataHandleIF<PMAPExchange<M,V>, double>
{
public:

    //Constructor
    PMAPExchange(const M& mapper, V& pmap) : mapper_(mapper), pmap_(pmap) {}

    bool contains(int dim, int codim) const
    {
        return (codim==0);
    }

    bool fixedsize(int dim, int codim) const
    {
        return false;
    }

    template <class EntityType>
    size_t size(EntityType& e) const
    {
        return (size_t)Pow(4,e.dimension + 1) * e.dimension;
    }

    template <class MessageBuffer, class EntityType>
    void gather( MessageBuffer& buff, const EntityType& e) const
    {
        int indexi = mapper_.index(e);
        for(int u=-1; u<=2; ++u){
            for(int v=-1; v<=2; ++v){
                for(int w=-1; w<=2; ++w){
                    for(int d=0; d<=3; ++d){
                        for(int k=0; k<=2; ++k){
                            buff.write( pmap_[indexi][u][v][w][d][k]);
                        }}}}}
    }

    template <class MessageBuffer, class EntityType>
    void scatter(MessageBuffer& buff, EntityType& e, size_t n)
    {
        assert( n == (size_t)( Pow(4,e.dimension + 1)*e.dimension ) );
        int indexi = mapper_.index(e);
        for(int u=-1; u<=2; ++u){
            for(int v=-1; v<=2; ++v){
                for(int w=-1; w<=2; ++w){
                    for(int d=0; d<=3; ++d){
                        for(int k=0; k<=2; ++k){
                            buff.read( pmap_[indexi][u][v][w][d][k]);
                        }}}}}
    }

private:
    const M& mapper_;
    V& pmap_;

};//end of PMAPExchange class


template <class M, class V>
class FLAGMAPExchange : public Dune::CommDataHandleIF<FLAGMAPExchange<M,V>, bool>
{
public:

    //Constructor
    FLAGMAPExchange(const M& mapper, V& flagmap) : mapper_(mapper), flagmap_(flagmap) {}

    bool contains(int dim, int codim) const
    {
        return (codim==0);
    }

    bool fixedsize(int dim, int codim) const
    {
        return false;
    }

    template <class EntityType>
    size_t size(EntityType& e) const
    {
        return (size_t)Pow(4,e.dimension);
    }

    template <class MessageBuffer, class EntityType>
    void gather( MessageBuffer& buff, const EntityType& e) const
    {
        int indexi = mapper_.index(e);
        for(int u=-1; u<=2; ++u){
            for(int v=-1; v<=2; ++v){
                for(int w=-1; w<=2; ++w){
                    buff.write( flagmap_[indexi][u][v][w]);
                }}}
    }

    template <class MessageBuffer, class EntityType>
    void scatter(MessageBuffer& buff, EntityType& e, size_t n)
    {
        assert( n == (size_t)( Pow(4,e.dimension) ) );
        int indexi = mapper_.index(e);
        for(int u=-1; u<=2; ++u){
            for(int v=-1; v<=2; ++v){
                for(int w=-1; w<=2; ++w){
                    buff.read( flagmap_[indexi][u][v][w]);
                }}}
    }

private:
    const M& mapper_;
    V& flagmap_;

};//end of FLAGMAPExchange class

#endif
