#include <config.h>

#include <iostream>
#include <iostream>
#include <iomanip>
#include <string>
#include <sys/stat.h>
#include <time.h>
#include <numeric>
#include <new>
#include <stdio.h>
#include <chrono>
#include <vector>
#include <dune/common/parallel/mpihelper.hh> // An initializer of MPI
#include <dune/common/exceptions.hh> // We use exceptions
#include "dune/common/parallel/mpihelper.hh"
#include <dune/alugrid/grid.hh> // We use the ALUGrid manager
#include <dune/grid/io/file/vtk/vtkwriter.hh>
#include <dune/grid/io/file/vtk/boundaryiterators.hh>
#include <dune/grid/common/scsgmapper.hh>
#include <dune/grid/common/mcmgmapper.hh>
#include <dune/grid/common/capabilities.hh>
#include <dune/grid/common/backuprestore.hh>
//#include <dune/subgrid/subgrid.hh>
#include <dune/grid/utility/persistentcontainer.hh>

#define IDIM 3

#include "sysTools.hh"
#include "gridpointer.hh"
#include "myparametrization.hh"
#include "deformation2.hh"
#include "facelayout.hh"



int main(int argc, char** argv)
{
    try{
        //Maybe initialize Mpi
        Dune::MPIHelper& helper = Dune::MPIHelper::instance(argc, argv);
        std::cout << "Hello World! This is intersectionBug." << std::endl;
        if(Dune::MPIHelper::isFake)
            std::cout<< "This is a sequential program." << std::endl;
        else
            std::cout<<"I am rank "<<helper.rank()<<" of "<<helper.size()
                     <<" processes!"<<std::endl;

        typedef Dune::ALUGrid<IDIM,IDIM,Dune::cube,Dune::nonconforming> GridType; 
        typedef GridType::ctype ct;
        const int dim = GridType::dimension;
        typedef Dune::FieldVector< ct,dim > DomainType;

        typedef typename GridType::LeafGridView LeafGridView;
        typedef Dune::PersistentContainer< GridType,MyParametrization<DomainType,DomainType > > ParametrizationsType;
        typedef Dune::PersistentContainer< GridType,int > NewsType;
        typedef Dune::LeafMultipleCodimMultipleGeomTypeMapper< GridType,Dune::MCMGElementLayout > LeafElementMapperType;
        typedef Dune::LeafMultipleCodimMultipleGeomTypeMapper< GridType,Dune::MCMGVertexLayout > LeafVertexMapperType;
        typedef Dune::LevelMultipleCodimMultipleGeomTypeMapper< GridType,Dune::MCMGFaceLayout> LevelFaceMapperType;
        typedef Dune::LevelMultipleCodimMultipleGeomTypeMapper< GridType,Dune::MCMGElementLayout> LevelElementMapperType;
        typedef std::vector<LevelElementMapperType*> LevelElementMapperPointerType;
        typedef std::vector<LevelFaceMapperType*> LevelFaceMapperPointerType;
        typedef DeformationDiscreteFunction2< DomainType, GridType, ParametrizationsType,Dune::MPIHelper, NewsType> DeformationType;
        typedef Dune::GeometryGrid<GridType,DeformationType> GeometryGridType;
        typedef Dune::LeafMultipleCodimMultipleGeomTypeMapper< GeometryGridType,Dune::MCMGElementLayout > CLeafElementMapperType;
        typedef GeometryGridType::LeafGridView CLeafGridView;
        typedef typename CLeafGridView::Codim<0>::Partition<Dune::InteriorBorder_Partition>::Iterator GEIterator;
        typedef typename LeafGridView::Codim<0>::Partition<Dune::InteriorBorder_Partition>::Iterator EIterator;
        typedef typename EIterator::Entity Element;
        typedef typename GEIterator::Entity GElement;
        typedef typename GElement::Geometry GEGeometry;
        typedef typename GEGeometry::GlobalCoordinate GGC;
        typedef typename CLeafGridView::IntersectionIterator GIIterator;
        typedef typename LeafGridView::IntersectionIterator IIterator;
        typedef typename IIterator::Intersection Intersection;
        typedef typename GIIterator::Intersection GIntersection;
    


        //----------------------//
        // CREATE OUTPUT FOLDER //
        //----------------------//
        if(helper.rank() == 0)
            std::cout << "Starting folder selection..." << std::endl;
        MPI_Barrier(helper.getCommunicator());

        std::string path("");
        std::string outputFolder = "outputs";
        std::string outputPath = "runOutputFolder";
        createFolder("",outputFolder,helper);
        createFolder(outputFolder,outputPath,helper);
        outputPath= outputFolder + "/" + outputPath;

        MPI_Barrier(helper.getCommunicator());
        if(helper.rank() == 0)
            std::cout << "End of folder selection..." << std::endl;
        MPI_Barrier(helper.getCommunicator());


        std::string inputGridPath = "grids/rampa15_visc2";
        GridPointer<GridType,0> gridPtr(inputGridPath, helper);
        GridType& grid = *gridPtr;
        gridPtr.loadBalance();
        Dune::MCMGElementLayout<dim> layout;
        LeafElementMapperType leafElementMapper(grid,layout);

        ParametrizationsType parametrizations(grid,0);
        NewsType news(grid,0);

        MPI_Barrier(helper.getCommunicator());
        if(helper.rank() == 0)
            std::cout << "Computing parametrizations..." << std::endl;
        MPI_Barrier(helper.getCommunicator());
        computeParametrizationsUNSTRUCTURED< DomainType >(gridPtr,leafElementMapper,parametrizations);
        DeformationType deformation(grid,parametrizations,helper,news);
        MPI_Barrier(helper.getCommunicator());
        if(helper.rank() == 0)
            std::cout << "End of parametrizations computation..." << std::endl;
        MPI_Barrier(helper.getCommunicator());

        GeometryGridType cgrid(grid,deformation);

        CLeafGridView cleafGridView = cgrid.leafGridView();

        //        typename Dune::LeafMultipleCodimMultipleGeomTypeMapper<GridType,Dune::MCMGElementLayout > leafElementMapper(grid,layout);

        typedef typename CLeafGridView::Codim<0>::Partition<Dune::InteriorBorder_Partition>::Iterator GEIterator;
        typedef typename GEIterator::Entity GElement;



        GEIterator endit = cleafGridView.end<0,Dune::InteriorBorder_Partition>();
        GEIterator beginit = cleafGridView.begin<0,Dune::InteriorBorder_Partition>();
        for(GEIterator it = beginit; it != endit; ++it)
            {
                const GElement & e = *it;
                const Element & he = e.impl().hostEntity();
                int l = e.level();
                //REMARK the next statement would give segmentation fault
                //                int indexi = cleafElementMapper.map(e);
                int indexi = leafElementMapper.index(he);

                if(indexi == 43) {
                    std::cout << "Element 43 GCorners" << std::endl;
                    for(int i = 0; i < 8; ++i){
                        GGC corner = e.geometry().corner(i);
                        std::cout << "gcorner " << i << " ";
                        for(int d = 0; d < 3; ++d)
                            std::cout << std::setprecision(16) << std::fixed << corner[d] << " ";
                        std::cout << std::endl;
                    }
                }

                GIIterator gendiit = cleafGridView.ibegin( e );
                GIIterator gbeginiit = cleafGridView.iend( e );
                for(GIIterator giit = gbeginiit; giit != gendiit; ++giit)
                    {
                        const GIntersection & in = *giit;
                        typename GIntersection::Geometry iGeo = in.geometry();
                        int indexInInside = giit->indexInInside();
                        if(indexi == 43 && indexInInside == 3){
                            std::cout << "Intersection 3 GCorners" << std::endl;
                            for(int c = 0; c < 4; ++c){
                                GGC corner = iGeo.corner(c);
                                std::cout << "gcorner " << c << " " << corner[0] << " " << corner[1]<< " " << corner[2] << std::endl;
                            }
                        }
                    }

                /////////////////////////////////////////////////////////////////7

                if(indexi == 43) {
                    std::cout << "Element 43 Corners" << std::endl;
                    for(int i = 0; i < 8; ++i){
                        GGC corner = he.geometry().corner(i);
                        std::cout << "corner " << i << " ";
                        for(int d = 0; d < 3; ++d)
                            std::cout << std::setprecision(16) << std::fixed << corner[d] << " ";
                        std::cout << std::endl;
                    }
                }


                const auto leafGridView = grid.leafGridView();
                IIterator endiit = leafGridView.ibegin( he );
                IIterator beginiit = leafGridView.iend( he );
                for(IIterator iit = beginiit; iit != endiit; ++iit)
                    {
                        const Intersection & in = *iit;
                        typename Intersection::Geometry iGeo = in.geometry();
                        int indexInInside = iit->indexInInside();
                        if(indexi == 43 && indexInInside == 3){
                            std::cout << "Intersection 3 Corners" << std::endl;
                            for(int c = 0; c < 4; ++c){
                                GGC corner = iGeo.corner(c);
                                std::cout << "corner " << c << " " << corner[0] << " " << corner[1]<< " " << corner[2] << std::endl;
                            }
                        }
                    }

            }




    }
    catch (Dune::Exception &e){
        std::cerr << "Dune reported error: " << e << std::endl;
    }
    catch (...){
        std::cerr << "Unknown exception thrown!" << std::endl;
    }
}
